class IOHyperModel(object):
    """A mixin class connecting the input nodes and heads with the adapters.

    This class is extended by the input nodes and the heads. The AutoModel calls the
    functions to get the corresponding adapters and pass the information back to the
    input nodes and heads.
    """

    def get_adapter(self):
        """Get the corresponding adapter.

        # Returns
            An instance of a subclass of autokeras.engine.Adapter.
        """
        raise NotImplementedError

    def config_from_adapter(self, adapter):
        """Load the learned information on dataset from the adapter.

        # Arguments
            adapter: An instance of a subclass of autokeras.engine.Adapter.
        """
        raise NotImplementedError



from typing import Optional
from typing import Tuple
from typing import Union

from tensorflow.keras.layers.experimental import preprocessing
from tensorflow.python.util import nest

from autokeras import adapters
from autokeras import keras_layers
from autokeras.engine import block as block_module


class Normalization(block_module.Block):
    """ Perform basic image transformation and augmentation.

    # Arguments
        axis: Integer or tuple of integers, the axis or axes that should be
            normalized (typically the features axis). We will normalize each element
            in the specified axis. The default is '-1' (the innermost axis); 0 (the
            batch axis) is not allowed.
    """

    def __init__(self, axis: int = -1, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        return preprocessing.Normalization(axis=self.axis)(input_node)

    def get_config(self):
        config = super().get_config()
        config.update({'axis': self.axis})
        return config


class TextToIntSequence(block_module.Block):
    """Convert raw texts to sequences of word indices.

    # Arguments
        output_sequence_length: Int. The maximum length of a sentence. If
            unspecified, it would be tuned automatically.
        max_tokens: Int. The maximum size of the vocabulary. Defaults to 20000.
    """

    def __init__(self,
                 output_sequence_length: Optional[int] = None,
                 max_tokens: int = 20000,
                 **kwargs):
        super().__init__(**kwargs)
        self.output_sequence_length = output_sequence_length
        self.max_tokens = max_tokens

    def get_config(self):
        config = super().get_config()
        config.update({
            'output_sequence_length': self.output_sequence_length,
            'max_tokens': self.max_tokens,
        })
        return config

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        if self.output_sequence_length is not None:
            output_sequence_length = self.output_sequence_length
        else:
            output_sequence_length = hp.Choice('output_sequence_length',
                                               [64, 128, 256, 512], default=64)
        output_node = preprocessing.TextVectorization(
            max_tokens=self.max_tokens,
            output_mode='int',
            output_sequence_length=output_sequence_length)(input_node)
        return output_node


class TextToNgramVector(block_module.Block):
    """Convert raw texts to n-gram vectors.

    # Arguments
        max_tokens: Int. The maximum size of the vocabulary. Defaults to 20000.
        ngrams: Int or tuple of ints. Passing an integer will create ngrams up to
            that integer, and passing a tuple of integers will create ngrams for the
            specified values in the tuple. If left unspecified, it will be tuned
            automatically.
    """

    def __init__(self,
                 max_tokens: int = 20000,
                 ngrams: Union[int, Tuple[int], None] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.max_tokens = max_tokens
        self.ngrams = ngrams

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        if self.ngrams is not None:
            ngrams = self.ngrams
        else:
            ngrams = hp.Int('ngrams', min_value=1, max_value=2, default=2)
        return preprocessing.TextVectorization(
            max_tokens=self.max_tokens,
            ngrams=ngrams,
            output_mode='tf-idf')(input_node)

    def get_config(self):
        config = super().get_config()
        config.update({
            'max_tokens': self.max_tokens,
            'ngrams': self.ngrams,
        })
        return config


class ImageAugmentation(block_module.Block):
    """Collection of various image augmentation methods.

    # Arguments
        translation_factor: A positive float represented as fraction value, or a
            tuple of 2 representing fraction for translation vertically and
            horizontally.  For instance, `translation_factor=0.2` result in a random
            translation factor within 20% of the width and height.
            If left unspecified, it will be tuned automatically.
        vertical_flip: Boolean. Whether to flip the image vertically.
            If left unspecified, it will be tuned automatically.
        horizontal_flip: Boolean. Whether to flip the image horizontally.
            If left unspecified, it will be tuned automatically.
        rotation_factor: Float. A positive float represented as fraction of 2pi
            upper bound for rotating clockwise and counter-clockwise. When
            represented as a single float, lower = upper.
            If left unspecified, it will be tuned automatically.
        zoom_factor: A positive float represented as fraction value, or a tuple of 2
            representing fraction for zooming vertically and horizontally. For
            instance, `zoom_factor=0.2` result in a random zoom factor from 80% to
            120%. If left unspecified, it will be tuned automatically.
        contrast_factor: A positive float represented as fraction of value, or a
            tuple of size 2 representing lower and upper bound. When represented as a
            single float, lower = upper. The contrast factor will be randomly picked
            between [1.0 - lower, 1.0 + upper]. If left unspecified, it will be tuned
            automatically.
    """

    def __init__(self,
                 translation_factor=None,
                 vertical_flip=None,
                 horizontal_flip=None,
                 rotation_factor=None,
                 zoom_factor=None,
                 contrast_factor=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.translation_factor = translation_factor
        self.horizontal_flip = horizontal_flip
        self.vertical_flip = vertical_flip
        self.rotation_factor = rotation_factor
        self.zoom_factor = zoom_factor
        self.contrast_factor = contrast_factor

    @staticmethod
    def _get_fraction_value(value):
        if isinstance(value, tuple):
            return value
        return value, value

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        output_node = input_node

        # Translate
        translation_factor = self.translation_factor
        if translation_factor is None:
            translation_factor = hp.Choice('translation_factor', [0.0, 0.1])
        if translation_factor != 0 and translation_factor != (0, 0):
            height_factor, width_factor = self._get_fraction_value(
                translation_factor)
            output_node = preprocessing.RandomTranslation(
                height_factor, width_factor)(output_node)

        # Flip
        horizontal_flip = self.horizontal_flip
        if horizontal_flip is None:
            horizontal_flip = hp.Boolean('horizontal_flip', default=True)
        vertical_flip = self.vertical_flip
        if self.vertical_flip is None:
            vertical_flip = hp.Boolean('vertical_flip', default=True)
        if not horizontal_flip and not vertical_flip:
            flip_mode = ''
        elif horizontal_flip and vertical_flip:
            flip_mode = 'horizontal_and_vertical'
        elif horizontal_flip and not vertical_flip:
            flip_mode = 'horizontal'
        elif not horizontal_flip and vertical_flip:
            flip_mode = 'vertical'
        if flip_mode != '':
            output_node = preprocessing.RandomFlip(
                mode=flip_mode)(output_node)

        # Rotate
        rotation_factor = self.rotation_factor
        if rotation_factor is None:
            rotation_factor = hp.Choice('rotation_factor', [0.0, 0.1])
        if rotation_factor != 0:
            output_node = preprocessing.RandomRotation(
                rotation_factor)(output_node)

        # Zoom
        zoom_factor = self.zoom_factor
        if zoom_factor is None:
            zoom_factor = hp.Choice('zoom_factor', [0.0, 0.1])
        if zoom_factor != 0 and zoom_factor != (0, 0):
            height_factor, width_factor = self._get_fraction_value(
                zoom_factor)
            # TODO: Add back RandomZoom when it is ready.
            # output_node = preprocessing.RandomZoom(
            # height_factor, width_factor)(output_node)

        # Contrast
        contrast_factor = self.contrast_factor
        if contrast_factor is None:
            contrast_factor = hp.Choice('contrast_factor', [0.0, 0.1])
        if contrast_factor != 0 and contrast_factor != (0, 0):
            output_node = preprocessing.RandomContrast(
                contrast_factor)(output_node)

        return output_node

    def get_config(self):
        config = super().get_config()
        config.update({
            'translation_factor': self.translation_factor,
            'horizontal_flip': self.horizontal_flip,
            'vertical_flip': self.vertical_flip,
            'rotation_factor': self.rotation_factor,
            'zoom_factor': self.zoom_factor,
            'contrast_factor': self.contrast_factor,
        })
        return config


class CategoricalToNumerical(block_module.Block):
    """Encode the categorical features to numerical features."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.column_types = None
        self.column_names = None

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        encoding = []
        for column_name in self.column_names:
            column_type = self.column_types[column_name]
            if column_type == adapters.CATEGORICAL:
                # TODO: Search to use one-hot or int.
                encoding.append(keras_layers.INT)
            else:
                encoding.append(keras_layers.NONE)
        return keras_layers.MultiColumnCategoricalEncoding(encoding)(input_node)

    @classmethod
    def from_config(cls, config):
        column_types = config.pop('column_types')
        column_names = config.pop('column_names')
        instance = cls(**config)
        instance.column_types = column_types
        instance.column_names = column_names
        return instance

    def get_config(self):
        config = super().get_config()
        config.update({'column_types': self.column_types,
                       'column_names': self.column_names})
        return config



import pandas as pd


class StructuredDataMixin(object):

    def check(self, column_names, column_types):
        if column_types:
            for column_type in column_types.values():
                if column_type not in ['categorical', 'numerical']:
                    raise ValueError(
                        'Column_types should be either "categorical" '
                        'or "numerical", but got {name}'.format(name=column_type))
        if column_names and column_types:
            for column_name in column_types:
                if column_name not in column_names:
                    raise ValueError('Column_names and column_types are '
                                     'mismatched. Cannot find column name '
                                     '{name} in the data.'.format(name=column_name))

    def read_for_predict(self, x):
        if isinstance(x, str):
            x = pd.read_csv(x)
            if self._target_col_name in x:
                x.pop(self._target_col_name)
        return x



import copy

import numpy as np
import pytest
import tensorflow as tf

from autokeras.adapters import input_adapter
from autokeras.utils import data_utils
from tests import utils


def test_structured_data_input_col_type_without_name():
    num_data = 500
    train_x = utils.generate_structured_data(num_data)
    with pytest.raises(ValueError) as info:
        adapter = input_adapter.StructuredDataInputAdapter(
            column_types=utils.COLUMN_TYPES_FROM_NUMPY)
        adapter.transform(train_x)
    assert str(info.value) == 'Column names must be specified.'


def test_structured_data_input_less_col_name():
    (x, _), _1 = utils.dataframe_numpy()
    with pytest.raises(ValueError) as info:
        adapter = input_adapter.StructuredDataInputAdapter(
            column_names=utils.LESS_COLUMN_NAMES_FROM_CSV)
        adapter.fit_transform(x)
    assert 'Expect column_names to have length' in str(info.value)


def test_structured_data_input_name_type_mismatch():
    (x, _), _1 = utils.dataframe_dataframe()
    column_types = copy.copy(utils.COLUMN_TYPES_FROM_CSV)
    column_types['age_'] = column_types.pop('age')
    with pytest.raises(ValueError) as info:
        adapter = input_adapter.StructuredDataInputAdapter(
            column_types=column_types)
        adapter.transform(x)
    assert 'Column_names and column_types are mismatched.' in str(info.value)


def test_structured_data_input_unsupported_type():
    x = 'unknown'
    with pytest.raises(TypeError) as info:
        adapter = input_adapter.StructuredDataInputAdapter(
            column_names=utils.COLUMN_TYPES_FROM_NUMPY,
            column_types=utils.COLUMN_TYPES_FROM_NUMPY)
        adapter.transform(x)
    assert 'Unsupported type' in str(info.value)


def test_structured_data_input_transform():
    (x, _), _1 = utils.dataframe_dataframe()
    adapter = input_adapter.StructuredDataInputAdapter()
    adapter.fit_transform(x)
    assert adapter.column_names[0] == 'sex'
    assert adapter.column_types == utils.COLUMN_TYPES_FROM_CSV


def test_structured_data_input_dataset():
    (x, _), _1 = utils.dataframe_dataframe()
    x = tf.data.Dataset.from_tensor_slices(x.to_numpy().astype(np.unicode))
    adapter = input_adapter.StructuredDataInputAdapter()
    x = adapter.fit_transform(x)
    assert isinstance(x, tf.data.Dataset)


def test_partial_column_types():
    adapter = input_adapter.StructuredDataInputAdapter(
        column_names=utils.COLUMN_NAMES_FROM_CSV,
        column_types=utils.PARTIAL_COLUMN_TYPES_FROM_CSV)
    (x, y), (val_x, val_y) = utils.dataframe_numpy()
    dataset = x.values.astype(np.unicode)
    adapter.transform(dataset)
    assert adapter.column_types['fare'] == 'categorical'


def test_image_input_adapter_transform_to_dataset():
    x = utils.generate_data()
    adapter = input_adapter.ImageInputAdapter()
    assert isinstance(adapter.transform(x), tf.data.Dataset)


def test_image_input_adapter_shape_is_list():
    x = utils.generate_data()
    adapter = input_adapter.ImageInputAdapter()
    adapter.fit_transform(x)
    assert isinstance(adapter.shape, list)
    assert all(map(lambda x: isinstance(x, int), adapter.shape))


def test_image_input_with_three_dim():
    x = utils.generate_data(shape=(32, 32))
    adapter = input_adapter.ImageInputAdapter()
    x = adapter.transform(x)
    assert isinstance(x, tf.data.Dataset)
    for a in x:
        assert a.shape[1:] == (32, 32, 1)
        break


def test_image_input_with_illegal_dim():
    x = utils.generate_data(shape=(32,))
    adapter = input_adapter.ImageInputAdapter()
    with pytest.raises(ValueError) as info:
        x = adapter.transform(x)
    assert 'Expect the data to ImageInput to have 3' in str(info.value)


def test_image_input_unsupported_type():
    x = 'unknown'
    adapter = input_adapter.ImageInputAdapter()
    with pytest.raises(TypeError) as info:
        x = adapter.transform(x)
    assert 'Expect the data to ImageInput to be numpy' in str(info.value)


def test_image_input_numerical():
    x = np.array([[['unknown']]])
    adapter = input_adapter.ImageInputAdapter()
    with pytest.raises(TypeError) as info:
        x = adapter.transform(x)
    assert 'Expect the data to ImageInput to be numerical' in str(info.value)


def test_input_type_error():
    x = 'unknown'
    adapter = input_adapter.InputAdapter()
    with pytest.raises(TypeError) as info:
        adapter.check(x)
        x = adapter.transform(x)
    assert 'Expect the data to Input to be numpy' in str(info.value)


def test_input_numerical():
    x = np.array([[['unknown']]])
    adapter = input_adapter.InputAdapter()
    with pytest.raises(TypeError) as info:
        adapter.check(x)
        x = adapter.transform(x)
    assert 'Expect the data to Input to be numerical' in str(info.value)


def test_text_dataset():
    x = tf.data.Dataset.from_tensor_slices(np.array([
        'a b c',
        'b b c',
    ]))
    adapter = input_adapter.TextInputAdapter()
    x = adapter.transform(x)
    assert data_utils.dataset_shape(x).as_list() == [None, 1]
    assert isinstance(x, tf.data.Dataset)


def test_text_dataset_batch():
    x = tf.data.Dataset.from_tensor_slices(np.array([
        'a b c',
        'b b c',
    ])).batch(32)
    adapter = input_adapter.TextInputAdapter()
    x = adapter.transform(x)
    assert data_utils.dataset_shape(x).as_list() == [None, 1]
    assert isinstance(x, tf.data.Dataset)


def test_text_np():
    x = np.array([
        'a b c',
        'b b c',
    ])
    adapter = input_adapter.TextInputAdapter()
    x = adapter.transform(x)
    assert data_utils.dataset_shape(x).as_list() == [None, 1]
    assert isinstance(x, tf.data.Dataset)


def test_text_input_type_error():
    x = 'unknown'
    adapter = input_adapter.TextInputAdapter()
    with pytest.raises(TypeError) as info:
        x = adapter.transform(x)
    assert 'Expect the data to TextInput to be numpy' in str(info.value)


def test_text_input_with_illegal_dim():
    x = utils.generate_data(shape=(32,))
    adapter = input_adapter.TextInputAdapter()
    with pytest.raises(ValueError) as info:
        x = adapter.transform(x)
    assert 'Expect the data to TextInput to have 1' in str(info.value)


def test_text_string():
    x = np.array([1, 2])
    adapter = input_adapter.TextInputAdapter()
    with pytest.raises(TypeError) as info:
        x = adapter.transform(x)
    assert 'Expect the data to TextInput to be strings' in str(info.value)


def test_time_series_input_type_error():
    x = 'unknown'
    adapter = input_adapter.TimeseriesInputAdapter(2)
    with pytest.raises(TypeError) as info:
        x = adapter.transform(x)
    assert 'Expect the data in TimeseriesInput to be numpy' in str(info.value)


def test_time_series_input_with_illegal_dim():
    x = utils.generate_data(shape=(32, 32))
    adapter = input_adapter.TimeseriesInputAdapter(2)
    with pytest.raises(ValueError) as info:
        x = adapter.transform(x)
    assert 'Expect the data in TimeseriesInput to have 2' in str(info.value)


def test_time_series_input_col_type_without_name():
    num_data = 500
    train_x = utils.generate_structured_data(num_data)
    with pytest.raises(ValueError) as info:
        adapter = input_adapter.TimeseriesInputAdapter(
            lookback=2,
            column_types=utils.COLUMN_TYPES_FROM_NUMPY)
        adapter.transform(train_x)
    assert str(info.value) == 'Column names must be specified.'


def test_time_series_input_less_col_name():
    (x, _), _1 = utils.dataframe_numpy()
    with pytest.raises(ValueError) as info:
        adapter = input_adapter.TimeseriesInputAdapter(
            lookback=2,
            column_names=utils.LESS_COLUMN_NAMES_FROM_CSV)
        adapter.transform(x)
    assert 'Expect column_names to have length' in str(info.value)


def test_time_series_input_name_type_mismatch():
    (x, _), _1 = utils.dataframe_dataframe()
    column_types = copy.copy(utils.COLUMN_TYPES_FROM_CSV)
    column_types['age_'] = column_types.pop('age')
    with pytest.raises(ValueError) as info:
        adapter = input_adapter.TimeseriesInputAdapter(
            lookback=2,
            column_types=column_types)
        adapter.transform(x)
    assert 'Column_names and column_types are mismatched.' in str(info.value)


def test_time_series_input_transform():
    x = utils.generate_data(shape=(32,))
    adapter = input_adapter.TimeseriesInputAdapter(2)
    x = adapter.transform(x)
    for row in x.as_numpy_iterator():
        assert row.ndim == 3



from typing import Optional

import tensorflow as tf
from tensorflow.keras import applications
from tensorflow.keras import layers
from tensorflow.python.util import nest

from autokeras.blocks import reduction
from autokeras.engine import block as block_module
from autokeras.utils import layer_utils
from autokeras.utils import utils

RESNET_V1 = {
    'resnet50': applications.ResNet50,
    'resnet101': applications.ResNet101,
    'resnet152': applications.ResNet152,
}

RESNET_V2 = {
    'resnet50_v2': applications.ResNet50V2,
    'resnet101_v2': applications.ResNet101V2,
    'resnet152_v2': applications.ResNet152V2,
}


class DenseBlock(block_module.Block):
    """Block for Dense layers.

    # Arguments
        num_layers: Int. The number of Dense layers in the block.
            If left unspecified, it will be tuned automatically.
        use_bn: Boolean. Whether to use BatchNormalization layers.
            If left unspecified, it will be tuned automatically.
        dropout: Float. The dropout rate for the layers.
            If left unspecified, it will be tuned automatically.
    """

    def __init__(self,
                 num_layers: Optional[int] = None,
                 use_batchnorm: Optional[bool] = None,
                 dropout: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.num_layers = num_layers
        self.use_batchnorm = use_batchnorm
        self.dropout = dropout

    def get_config(self):
        config = super().get_config()
        config.update({
            'num_layers': self.num_layers,
            'use_batchnorm': self.use_batchnorm,
            'dropout': self.dropout})
        return config

    def build(self, hp, inputs=None):
        inputs = nest.flatten(inputs)
        utils.validate_num_inputs(inputs, 1)
        input_node = inputs[0]
        output_node = input_node
        output_node = reduction.Flatten().build(hp, output_node)

        num_layers = self.num_layers or hp.Choice('num_layers', [1, 2, 3], default=2)
        use_batchnorm = self.use_batchnorm
        if use_batchnorm is None:
            use_batchnorm = hp.Boolean('use_batchnorm', default=False)
        if self.dropout is not None:
            dropout = self.dropout
        else:
            dropout = hp.Choice('dropout', [0.0, 0.25, 0.5], default=0)

        for i in range(num_layers):
            units = hp.Choice(
                'units_{i}'.format(i=i),
                [16, 32, 64, 128, 256, 512, 1024],
                default=32)
            output_node = layers.Dense(units)(output_node)
            if use_batchnorm:
                output_node = layers.BatchNormalization()(output_node)
            output_node = layers.ReLU()(output_node)
            if dropout > 0:
                output_node = layers.Dropout(dropout)(output_node)
        return output_node


class RNNBlock(block_module.Block):
    """An RNN Block.

    # Arguments
        return_sequences: Boolean. Whether to return the last output in the
            output sequence, or the full sequence. Defaults to False.
        bidirectional: Boolean. Bidirectional RNN. If left unspecified, it will be
            tuned automatically.
        num_layers: Int. The number of layers in RNN. If left unspecified, it will
            be tuned automatically.
        layer_type: String. 'gru' or 'lstm'. If left unspecified, it will be tuned
            automatically.
    """

    def __init__(self,
                 return_sequences: bool = False,
                 bidirectional: Optional[bool] = None,
                 num_layers: Optional[int] = None,
                 layer_type: Optional[int] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.return_sequences = return_sequences
        self.bidirectional = bidirectional
        self.num_layers = num_layers
        self.layer_type = layer_type

    def get_config(self):
        config = super().get_config()
        config.update({
            'return_sequences': self.return_sequences,
            'bidirectional': self.bidirectional,
            'num_layers': self.num_layers,
            'layer_type': self.layer_type})
        return config

    def build(self, hp, inputs=None):
        inputs = nest.flatten(inputs)
        utils.validate_num_inputs(inputs, 1)
        input_node = inputs[0]
        shape = input_node.shape.as_list()
        if len(shape) != 3:
            raise ValueError(
                'Expect the input tensor to have '
                'at least 3 dimensions for rnn models, '
                'but got {shape}'.format(shape=input_node.shape))

        feature_size = shape[-1]
        output_node = input_node

        bidirectional = self.bidirectional
        if bidirectional is None:
            bidirectional = hp.Boolean('bidirectional', default=True)
        layer_type = self.layer_type or hp.Choice('layer_type',
                                                  ['gru', 'lstm'],
                                                  default='lstm')
        num_layers = self.num_layers or hp.Choice('num_layers',
                                                  [1, 2, 3],
                                                  default=2)
        rnn_layers = {
            'gru': layers.GRU,
            'lstm': layers.LSTM
        }
        in_layer = rnn_layers[layer_type]
        for i in range(num_layers):
            return_sequences = True
            if i == num_layers - 1:
                return_sequences = self.return_sequences
            if bidirectional:
                output_node = layers.Bidirectional(
                    in_layer(feature_size,
                             return_sequences=return_sequences))(output_node)
            else:
                output_node = in_layer(
                    feature_size,
                    return_sequences=return_sequences)(output_node)
        return output_node


class ConvBlock(block_module.Block):
    """Block for vanilla ConvNets.

    # Arguments
        kernel_size: Int. If left unspecified, it will be tuned automatically.
        num_blocks: Int. The number of conv blocks, each of which may contain
            convolutional, max pooling, dropout, and activation. If left unspecified,
            it will be tuned automatically.
        num_layers: Int. The number of convolutional layers in each block. If left
            unspecified, it will be tuned automatically.
        max_pooling: Boolean. Whether to use max pooling layer in each block. If left
            unspecified, it will be tuned automatically.
        separable: Boolean. Whether to use separable conv layers.
            If left unspecified, it will be tuned automatically.
        dropout: Float. Between 0 and 1. The dropout rate for after the
            convolutional layers. If left unspecified, it will be tuned
            automatically.
    """

    def __init__(self,
                 kernel_size: Optional[int] = None,
                 num_blocks: Optional[int] = None,
                 num_layers: Optional[int] = None,
                 max_pooling: Optional[bool] = None,
                 separable: Optional[bool] = None,
                 dropout: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.kernel_size = kernel_size
        self.num_blocks = num_blocks
        self.num_layers = num_layers
        self.max_pooling = max_pooling
        self.separable = separable
        self.dropout = dropout

    def get_config(self):
        config = super().get_config()
        config.update({
            'kernel_size': self.kernel_size,
            'num_blocks': self.num_blocks,
            'num_layers': self.num_layers,
            'max_pooling': self.max_pooling,
            'separable': self.separable,
            'dropout': self.dropout})
        return config

    def build(self, hp, inputs=None):
        inputs = nest.flatten(inputs)
        utils.validate_num_inputs(inputs, 1)
        input_node = inputs[0]
        output_node = input_node

        kernel_size = self.kernel_size or hp.Choice('kernel_size',
                                                    [3, 5, 7],
                                                    default=3)
        num_blocks = self.num_blocks or hp.Choice('num_blocks',
                                                  [1, 2, 3],
                                                  default=2)
        num_layers = self.num_layers or hp.Choice('num_layers',
                                                  [1, 2],
                                                  default=2)
        separable = self.separable
        if separable is None:
            separable = hp.Boolean('separable', default=False)

        if separable:
            conv = layer_utils.get_sep_conv(input_node.shape)
        else:
            conv = layer_utils.get_conv(input_node.shape)

        max_pooling = self.max_pooling
        if max_pooling is None:
            max_pooling = hp.Boolean('max_pooling', default=True)
        pool = layer_utils.get_max_pooling(input_node.shape)

        if self.dropout is not None:
            dropout = self.dropout
        else:
            dropout = hp.Choice('dropout', [0.0, 0.25, 0.5], default=0)

        for i in range(num_blocks):
            for j in range(num_layers):
                output_node = conv(
                    hp.Choice('filters_{i}_{j}'.format(i=i, j=j),
                              [16, 32, 64, 128, 256, 512],
                              default=32),
                    kernel_size,
                    padding=self._get_padding(kernel_size, output_node),
                    activation='relu')(output_node)
            if max_pooling:
                output_node = pool(
                    kernel_size - 1,
                    padding=self._get_padding(kernel_size - 1,
                                              output_node))(output_node)
            if dropout > 0:
                output_node = layers.Dropout(dropout)(output_node)
        return output_node

    @staticmethod
    def _get_padding(kernel_size, output_node):
        if all([kernel_size * 2 <= length
                for length in output_node.shape[1:-1]]):
            return 'valid'
        return 'same'


class MultiHeadSelfAttention(block_module.Block):
    """Block for Multi-Head Self-Attention.

    # Arguments
        head_size: Int. Dimensionality of the `query`, `key` and `value` tensors
            after the linear transformation. If left unspecified, it will be
            tuned automatically.
        num_heads: Int. The number of attention heads. If left unspecified,
            it will be tuned automatically.
    """

    def __init__(self,
                 head_size: Optional[int] = None,
                 num_heads: Optional[int] = 8,
                 **kwargs):
        super().__init__(**kwargs)
        self.head_size = head_size
        self.num_heads = num_heads

    def get_config(self):
        config = super().get_config()
        config.update({
            'head_size': self.head_size,
            'num_heads': self.num_heads})
        return config

    def build(self, hp, inputs=None):
        """
        # Arguments
             hp: HyperParameters. The hyperparameters for building the model.
             inputs: Tensor of Shape [batch_size, seq_len, embedding_dim]

        # Returns
            Self-Attention outputs of shape `[batch_size, seq_len, embedding_dim]`.
        """
        inputs = nest.flatten(inputs)
        utils.validate_num_inputs(inputs, 1)
        input_node = inputs[0]
        shape = input_node.shape.as_list()
        if len(shape) != 3:
            raise ValueError(
                'Expect the input tensor to have '
                '3 dimensions for multi-head self-attention, '
                'but got {shape}'.format(shape=input_node.shape))
        # input.shape = [batch_size, seq_len, embedding_dim]
        head_size = self.head_size or hp.Choice(
            'head_size',
            [32, 64, 128, 256, 512],
            default=128)
        num_heads = self.num_heads
        if num_heads is None:
            num_heads = 8

        if head_size % num_heads != 0:  # how to evaluate this condition
            raise ValueError(
                f"embedding dimension = {head_size} should be "
                f"divisible by number of heads = {num_heads}"
            )
        projection_dim = head_size // num_heads
        query_dense = layers.Dense(head_size)
        key_dense = layers.Dense(head_size)
        value_dense = layers.Dense(head_size)
        combine_heads = layers.Dense(head_size)
        batch_size = tf.shape(input_node)[0]
        query = query_dense(input_node)  # (batch_size, seq_len, head_size)
        key = key_dense(input_node)  # (batch_size, seq_len, head_size)
        value = value_dense(input_node)  # (batch_size, seq_len, head_size)
        query, key, value = [self.separate_heads(
            var, batch_size, num_heads,  projection_dim
        ) for var in [query, key, value]]
        attention, weights = self.attention(query, key, value)
        attention = tf.transpose(
            attention, perm=[0, 2, 1, 3]
        )  # (batch_size, seq_len, num_heads, projection_dim)
        concat_attention = tf.reshape(
            attention, (batch_size, tf.shape(attention)[1], self.head_size)
        )  # (batch_size, seq_len, head_size)
        output = combine_heads(
            concat_attention
        )  # (batch_size, seq_len, head_size)
        return output

    @staticmethod
    def attention(query, key, value):
        score = tf.matmul(query, key, transpose_b=True)
        dim_key = tf.cast(tf.shape(key)[-1], tf.float32)
        scaled_score = score / tf.math.sqrt(dim_key)
        weights = tf.nn.softmax(scaled_score, axis=-1)
        output = tf.matmul(weights, value)
        return output, weights

    @staticmethod
    def separate_heads(x, batch_size, num_heads, projection_dim):
        x = tf.reshape(x, (batch_size, -1, num_heads, projection_dim))
        return tf.transpose(x, perm=[0, 2, 1, 3])


class Transformer(block_module.Block):
    """Block for Transformer.
    The input should be tokenized sequences with the same length, where each element
    of a sequence should be the index of the word.

    # Example
    ```python
        # Using the Transformer Block with AutoModel.
        import autokeras as ak
        from tensorflow.keras import losses
        text_input = ak.TextInput()
        output_node = ak.TextToIntSequence(output_sequence_length=200)(text_input)
        output_node = ak.Transformer(embedding_dim=32,
                             pretraining='none',
                             num_heads=2,
                             dense_dim=32,
                             dropout = 0.25)(output_node)
        output_node = ak.SpatialReduction(reduction_type='global_avg')(output_node)
        output_node = ak.DenseBlock(num_layers=1, use_batchnorm = False)(output_node)
        output_node = ak.ClassificationHead(
            loss=losses.SparseCategoricalCrossentropy(),
            dropout = 0.25)(output_node)
        clf = ak.AutoModel(inputs=text_input, outputs=output_node, max_trials=2)
    ```
    # Arguments
        max_features: Int. Size of the vocabulary. Must be set if not using
            TextToIntSequence before this block. Defaults to 20001.
        pretraining: String. 'random' (use random weights instead any pretrained
            model), 'glove', 'fasttext' or 'word2vec'. Use pretrained word embedding.
            If left unspecified, it will be tuned automatically.
        embedding_dim: Int. Output dimension of the Attention block.
            If left unspecified, it will be tuned automatically.
        num_heads: Int. The number of attention heads. If left unspecified,
            it will be tuned automatically.
        dense_dim: Int. The output dimension of the Feed-Forward Network. If left
            unspecified, it will be tuned automatically.
        dropout: Float. Between 0 and 1. If left unspecified, it will be
            tuned automatically.
    """

    def __init__(self,
                 max_features: int = 20001,
                 pretraining: Optional[str] = None,
                 embedding_dim: Optional[int] = None,
                 num_heads: Optional[int] = None,
                 dense_dim: Optional[int] = None,
                 dropout: Optional[int] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.max_features = max_features
        self.pretraining = pretraining
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self. dense_dim = dense_dim
        self.dropout = dropout

    def get_config(self):
        config = super().get_config()
        config.update({
            'max_features': self.max_features,
            'pretraining': self.pretraining,
            'embedding_dim': self.embedding_dim,
            'num_heads': self.num_heads,
            'dense_dim': self.dense_dim,
            'dropout': self.dropout})
        return config

    def build(self, hp, inputs=None):
        """
        # Arguments
             hp: HyperParameters. The hyperparameters for building the model.
             inputs: Tensor of Shape [batch_size, seq_len]

        # Returns
            Output Tensor of shape `[batch_size, seq_len, embedding_dim]`.
        """
        inputs = nest.flatten(inputs)
        utils.validate_num_inputs(inputs, 1)
        pretraining = self.pretraining or hp.Choice(
            'pretraining',
            ['random', 'glove', 'fasttext', 'word2vec', 'none'],
            default='none')
        embedding_dim = self.embedding_dim or hp.Choice(
            'embedding_dim',
            [32, 64, 128, 256, 512],
            default=128)
        num_heads = self.num_heads or hp.Choice('num_heads', [8, 16, 32], default=8)

        dense_dim = self.dense_dim or hp.Choice('dense_dim',
                                                [128, 256, 512, 1024, 2048],
                                                default=2048)
        dropout = self.dropout or hp.Choice('dropout',
                                            [0.0, 0.25, 0.5],
                                            default=0)

        ffn = tf.keras.Sequential(
            [layers.Dense(dense_dim, activation="relu"),
             layers.Dense(embedding_dim), ]
        )

        layernorm1 = layers.LayerNormalization(epsilon=1e-6)
        layernorm2 = layers.LayerNormalization(epsilon=1e-6)
        dropout1 = layers.Dropout(dropout)
        dropout2 = layers.Dropout(dropout)
        # Token and Position Embeddings
        input_node = nest.flatten(inputs)[0]
        token_embedding = Embedding(max_features=self.max_features,
                                    pretraining=pretraining,
                                    embedding_dim=embedding_dim,
                                    dropout=dropout).build(hp, input_node)
        maxlen = input_node.shape[-1]
        batch_size = tf.shape(input_node)[0]
        positions = self.pos_array_funct(maxlen, batch_size)
        position_embedding = Embedding(max_features=maxlen,
                                       pretraining=pretraining,
                                       embedding_dim=embedding_dim,
                                       dropout=dropout).build(hp,
                                                              positions)
        output_node = tf.keras.layers.Add()([token_embedding,
                                             position_embedding])
        attn_output = MultiHeadSelfAttention(
            embedding_dim, num_heads).build(hp, output_node)
        attn_output = dropout1(attn_output)
        add_inputs_1 = tf.keras.layers.Add()([output_node, attn_output])
        out1 = layernorm1(add_inputs_1)
        ffn_output = ffn(out1)
        ffn_output = dropout2(ffn_output)
        add_inputs_2 = tf.keras.layers.Add()([out1, ffn_output])
        output = layernorm2(add_inputs_2)
        return output

    @staticmethod
    def pos_array_funct(maxlen, batch_size):
        pos_ones = tf.ones((batch_size, 1), dtype=tf.int32)
        positions = tf.range(start=0, limit=maxlen, delta=1)
        positions = tf.expand_dims(positions, 0)
        positions = tf.matmul(pos_ones, positions)
        return positions


class KerasApplicationBlock(block_module.Block):
    """Blocks extending Keras applications."""

    def __init__(self,
                 pretrained,
                 models,
                 min_size,
                 **kwargs):
        super().__init__(**kwargs)
        self.pretrained = pretrained
        self.models = models
        self.min_size = min_size

    def get_config(self):
        config = super().get_config()
        config.update({
            'pretrained': self.pretrained,
        })
        return config

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]

        pretrained = self.pretrained
        if input_node.shape[3] not in [1, 3]:
            if self.pretrained:
                raise ValueError(
                    'When pretrained is set to True, expect input to '
                    'have 1 or 3 channels, bug got '
                    '{channels}.'.format(channels=input_node.shape[3]))
            pretrained = False
        if pretrained is None:
            pretrained = hp.Boolean('pretrained', default=False)

        if len(self.models) > 1:
            version = hp.Choice('version', list(self.models.keys()))
        else:
            version = list(self.models.keys())[0]

        min_size = self.min_size
        if hp.Boolean('imagenet_size', default=False):
            min_size = 224
        if (input_node.shape[1] < min_size or
                input_node.shape[2] < min_size):
            input_node = layers.experimental.preprocessing.Resizing(
                max(min_size, input_node.shape[1]),
                max(min_size, input_node.shape[2])
            )(input_node)
        if input_node.shape[3] == 1:
            input_node = layers.Concatenate()([input_node] * 3)
        if input_node.shape[3] != 3:
            input_node = layers.Conv2D(filters=3,
                                       kernel_size=1,
                                       padding='same')(input_node)

        if pretrained:
            model = self.models[version](
                weights='imagenet',
                include_top=False)
            model.trainable = hp.Boolean('trainable', default=False)
        else:
            model = self.models[version](
                weights=None,
                include_top=False,
                input_shape=input_node.shape[1:])

        return model(input_node)


class ResNetBlock(KerasApplicationBlock):
    """Block for ResNet.

    # Arguments
        version: String. 'v1', 'v2'. The type of ResNet to use.
            If left unspecified, it will be tuned automatically.
        pretrained: Boolean. Whether to use ImageNet pretrained weights.
            If left unspecified, it will be tuned automatically.
    """

    def __init__(self,
                 version: Optional[str] = None,
                 pretrained: Optional[bool] = None,
                 **kwargs):
        if version is None:
            models = {**RESNET_V1, **RESNET_V2}
        elif version == 'v1':
            models = RESNET_V1
        elif version == 'v2':
            models = RESNET_V2
        else:
            raise ValueError(
                'Expect version to be "v1", or "v2", but got '
                '{version}.'.format(version=version))
        super().__init__(pretrained=pretrained,
                         models=models,
                         min_size=32,
                         **kwargs)
        self.version = version

    def get_config(self):
        config = super().get_config()
        config.update({
            'version': self.version,
        })
        return config


class XceptionBlock(KerasApplicationBlock):
    """Block for XceptionNet.

    An Xception structure, used for specifying your model with specific datasets.

    The original Xception architecture is from https://arxiv.org/abs/1610.02357.
    The data first goes through the entry flow, then through the middle flow which
    is repeated eight times, and finally through the exit flow.

    This XceptionBlock returns a similar architecture as Xception except without
    the last (optional) fully connected layer(s) and logistic regression.
    The size of this architecture could be decided by `HyperParameters`, to get an
    architecture with a half, an identical, or a double size of the original one.

    # Arguments
        pretrained: Boolean. Whether to use ImageNet pretrained weights.
            If left unspecified, it will be tuned automatically.
    """

    def __init__(self,
                 pretrained: Optional[bool] = None,
                 **kwargs):
        super().__init__(pretrained=pretrained,
                         models={'xception': applications.Xception},
                         min_size=71,
                         **kwargs)


class Embedding(block_module.Block):
    """Word embedding block for sequences.

    The input should be tokenized sequences with the same length, where each element
    of a sequence should be the index of the word.

    # Arguments
        max_features: Int. Size of the vocabulary. Must be set if not using
            TextToIntSequence before this block. Defaults to 20001.
        pretraining: String. 'random' (use random weights instead any pretrained
            model), 'glove', 'fasttext' or 'word2vec'. Use pretrained word embedding.
            If left unspecified, it will be tuned automatically.
        embedding_dim: Int. If left unspecified, it will be tuned automatically.
        dropout: Float. The dropout rate for after the Embedding layer.
            If left unspecified, it will be tuned automatically.
    """

    def __init__(self,
                 max_features: int = 20001,
                 pretraining: Optional[str] = None,
                 embedding_dim: Optional[int] = None,
                 dropout: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.max_features = max_features
        self.pretraining = pretraining
        self.embedding_dim = embedding_dim
        self.dropout = dropout

    def get_config(self):
        config = super().get_config()
        config.update({
            'max_features': self.max_features,
            'pretraining': self.pretraining,
            'embedding_dim': self.embedding_dim,
            'dropout': self.dropout})
        return config

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        # TODO: support more pretrained embedding layers.
        # glove, fasttext, and word2vec
        pretraining = self.pretraining or hp.Choice(
            'pretraining',
            ['random', 'glove', 'fasttext', 'word2vec', 'none'],
            default='none')
        embedding_dim = self.embedding_dim or hp.Choice(
            'embedding_dim',
            [32, 64, 128, 256, 512],
            default=128)
        if pretraining != 'none':
            # TODO: load from pretrained weights
            layer = layers.Embedding(
                input_dim=self.max_features,
                output_dim=embedding_dim,
                input_length=input_node.shape[1])
            # trainable=False,
            # weights=[embedding_matrix])
        else:
            layer = layers.Embedding(
                input_dim=self.max_features,
                output_dim=embedding_dim)
            # input_length=input_node.shape[1],
            # trainable=True)
        output_node = layer(input_node)
        if self.dropout is not None:
            dropout = self.dropout
        else:
            dropout = hp.Choice('dropout', [0.0, 0.25, 0.5], default=0.25)
        if dropout > 0:
            output_node = layers.Dropout(dropout)(output_node)
        return output_node



import numpy as np
import tensorflow as tf

from autokeras.engine import serializable
from autokeras.utils import data_utils


class Adapter(serializable.Serializable):
    """Adpat the input and output format for Keras Model.

    Adapter is used by the input nodes and the heads of the hypermodel. It analyzes
    the training data to get useful information, e.g., the shape of the data, which
    is required for building the Keras Model. It also converts the dataset to
    tf.data.Dataset format.

    # Arguments
        shape: Tuple of int. The input or output shape of the hypermodel.
        batch_size: Int. Number of samples per gradient update. Defaults to 32.
    """

    def __init__(self, shape=None, batch_size=32):
        self.shape = shape
        self.batch_size = batch_size

    def check(self, dataset):
        """Check if the dataset is valid for the input node.

        # Arguments
            dataset: Usually numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                The dataset to be checked.

        # Returns
            Boolean. Whether the dataset is valid for the input node.
        """
        return True

    def convert_to_dataset(self, dataset):
        """Convert supported formats of datasets to tf.data.Dataset.

        # Arguments
            dataset: Usually numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                The dataset to be converted.

        # Returns
            tf.data.Dataset. The converted dataset.
        """
        if isinstance(dataset, np.ndarray):
            dataset = tf.data.Dataset.from_tensor_slices(dataset)
        return data_utils.batch_dataset(dataset, self.batch_size)

    def fit(self, dataset):
        """Analyze the dataset and record useful information.

        # Arguments
            dataset: tf.data.Dataset.
        """
        self._record_dataset_shape(dataset)

    def fit_before_convert(self, dataset):
        """Analyze the dataset before converting to tf.data.Dataset.

        # Arguments
            dataset: Usually numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
        """
        pass

    def fit_transform(self, dataset):
        self.check(dataset)
        self.fit_before_convert(dataset)
        dataset = self.convert_to_dataset(dataset)
        self.fit(dataset)
        return dataset

    def _record_dataset_shape(self, dataset):
        self.shape = data_utils.dataset_shape(dataset)[1:].as_list()

    def transform(self, dataset):
        """Transform the input dataset to tf.data.Dataset.

        # Arguments
            dataset: Usually numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                The dataset to be transformed.

        # Returns
            tf.data.Dataset. The converted dataset.
        """
        self.check(dataset)
        return self.convert_to_dataset(dataset)

    def get_config(self):
        return {'shape': self.shape}



from typing import Callable
from typing import Dict
from typing import List
from typing import Union

import numpy as np
import tensorflow as tf
from tensorflow.keras.losses import Loss
from tensorflow.keras.metrics import Metric

DatasetType = Union[np.ndarray, tf.data.Dataset]
LossType = Union[str, Callable, Loss]
AcceptableMetric = Union[str, Callable, Metric]
MetricsType = Union[List[AcceptableMetric],
                    List[List[AcceptableMetric]],
                    Dict[str, AcceptableMetric]]



from unittest import mock

import tensorflow as tf
from tensorflow.keras.layers.experimental import preprocessing

from autokeras.engine import tuner as tuner_module
from autokeras.tuners import greedy
from tests import utils


def called_with_early_stopping(func):
    callbacks = func.call_args_list[0][1]['callbacks']
    return any([isinstance(callback, tf.keras.callbacks.EarlyStopping)
                for callback in callbacks])


@mock.patch('kerastuner.engine.base_tuner.BaseTuner.search')
@mock.patch('autokeras.engine.tuner.AutoTuner.final_fit')
def test_final_fit_with_specified_epochs(
        final_fit, super_search, tmp_path):
    tuner = greedy.Greedy(
        hypermodel=utils.build_graph(),
        directory=tmp_path)

    tuner.search(x=None, epochs=10)

    assert final_fit.call_args_list[0][1]['epochs'] == 10


@mock.patch('kerastuner.engine.base_tuner.BaseTuner.search')
@mock.patch('autokeras.engine.tuner.AutoTuner.final_fit')
def test_tuner_call_super_with_early_stopping(
        final_fit, super_search, tmp_path):
    tuner = greedy.Greedy(
        hypermodel=utils.build_graph(),
        directory=tmp_path)

    tuner.search(x=None, epochs=10)

    assert called_with_early_stopping(super_search)


@mock.patch('kerastuner.engine.base_tuner.BaseTuner.search')
@mock.patch('autokeras.engine.tuner.AutoTuner.final_fit')
@mock.patch('autokeras.engine.tuner.AutoTuner.get_best_models',
            return_value=[mock.Mock()])
def test_no_final_fit_without_epochs_and_fov(
        get_best_models, final_fit, super_search, tmp_path):
    tuner = greedy.Greedy(
        hypermodel=utils.build_graph(),
        directory=tmp_path)

    tuner.search(x=None, epochs=None, fit_on_val_data=False)

    final_fit.assert_not_called()


@mock.patch('kerastuner.engine.base_tuner.BaseTuner.search')
@mock.patch('autokeras.engine.tuner.AutoTuner.final_fit')
@mock.patch('autokeras.engine.tuner.AutoTuner._get_best_trial_epochs',
            return_value=2)
def test_final_fit_best_epochs_if_epoch_unspecified(
        best_epochs, final_fit, super_search, tmp_path):
    tuner = greedy.Greedy(
        hypermodel=utils.build_graph(),
        directory=tmp_path)

    tuner.search(x=mock.Mock(),
                 epochs=None,
                 fit_on_val_data=True,
                 validation_data=mock.Mock())

    assert final_fit.call_args_list[0][1]['epochs'] == 2


@mock.patch('kerastuner.engine.base_tuner.BaseTuner.search')
@mock.patch('autokeras.engine.tuner.AutoTuner.final_fit')
@mock.patch('autokeras.engine.tuner.AutoTuner._get_best_trial_epochs',
            return_value=2)
def test_super_with_1k_epochs_if_epoch_unspecified(
        best_epochs, final_fit, super_search, tmp_path):
    tuner = greedy.Greedy(
        hypermodel=utils.build_graph(),
        directory=tmp_path)

    tuner.search(x=mock.Mock(),
                 epochs=None,
                 fit_on_val_data=True,
                 validation_data=mock.Mock())

    assert super_search.call_args_list[0][1]['epochs'] == 1000
    assert called_with_early_stopping(super_search)


@mock.patch('kerastuner.engine.base_tuner.BaseTuner.search')
@mock.patch('autokeras.engine.tuner.AutoTuner.final_fit')
def test_tuner_not_call_super_search_with_overwrite(
        final_fit, super_search, tmp_path):
    tuner = greedy.Greedy(
        hypermodel=utils.build_graph(),
        directory=tmp_path)

    tuner.search(epochs=10)
    tuner.save()
    super_search.reset_mock()

    tuner = greedy.Greedy(
        hypermodel=utils.build_graph(),
        directory=tmp_path)
    tuner.search(epochs=10)

    super_search.assert_not_called()


def test_preprocessing_adapt():
    class MockLayer(preprocessing.TextVectorization):
        def adapt(self, *args, **kwargs):
            super().adapt(*args, **kwargs)
            self.is_called = True

    (x_train, y_train), (x_test, y_test) = utils.imdb_raw()
    dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train)).batch(32)
    layer1 = MockLayer(
        max_tokens=5000,
        output_mode='int',
        output_sequence_length=40)
    model = tf.keras.models.Sequential()
    model.add(tf.keras.Input(shape=(1,), dtype=tf.string))
    model.add(layer1)
    model.add(tf.keras.layers.Embedding(50001, 10))
    model.add(tf.keras.layers.Dense(1))

    tuner_module.AutoTuner.adapt(model, dataset)

    assert layer1.is_called



import tensorflow as tf


def get_global_average_pooling(shape):
    return [tf.keras.layers.GlobalAveragePooling1D,
            tf.keras.layers.GlobalAveragePooling2D,
            tf.keras.layers.GlobalAveragePooling3D][len(shape) - 3]


def get_global_max_pooling(shape):
    return [tf.keras.layers.GlobalMaxPool1D,
            tf.keras.layers.GlobalMaxPool2D,
            tf.keras.layers.GlobalMaxPool3D][len(shape) - 3]


def get_max_pooling(shape):
    return [tf.keras.layers.MaxPool1D,
            tf.keras.layers.MaxPool2D,
            tf.keras.layers.MaxPool3D][len(shape) - 3]


def get_conv(shape):
    return [tf.keras.layers.Conv1D,
            tf.keras.layers.Conv2D,
            tf.keras.layers.Conv3D][len(shape) - 3]


def get_sep_conv(shape):
    return [tf.keras.layers.SeparableConv1D,
            tf.keras.layers.SeparableConv2D,
            tf.keras.layers.Conv3D][len(shape) - 3]


def get_dropout(shape):
    return [tf.keras.layers.SpatialDropout1D,
            tf.keras.layers.SpatialDropout2D,
            tf.keras.layers.SpatialDropout3D][len(shape) - 3]



import os

import kerastuner
import pytest
from kerastuner.engine import hyperparameters as hp_module

import autokeras as ak
from autokeras import graph as graph_module


def test_set_hp():
    input_node = ak.Input((32,))
    output_node = input_node
    output_node = ak.DenseBlock()(output_node)
    head = ak.RegressionHead()
    head.output_shape = (1,)
    output_node = head(output_node)

    graph = graph_module.Graph(
        inputs=input_node,
        outputs=output_node,
        override_hps=[hp_module.Choice('dense_block_1/num_layers', [6], default=6)])
    hp = kerastuner.HyperParameters()
    graph.build(hp)

    for single_hp in hp.space:
        if single_hp.name == 'dense_block_1/num_layers':
            assert len(single_hp.values) == 1
            assert single_hp.values[0] == 6
            return
    assert False


def test_input_output_disconnect():
    input_node1 = ak.Input()
    output_node = input_node1
    _ = ak.DenseBlock()(output_node)

    input_node = ak.Input()
    output_node = input_node
    output_node = ak.DenseBlock()(output_node)
    output_node = ak.RegressionHead()(output_node)

    with pytest.raises(ValueError) as info:
        graph_module.Graph(inputs=input_node1, outputs=output_node)
    assert 'Inputs and outputs not connected.' in str(info.value)


def test_hyper_graph_cycle():
    input_node1 = ak.Input()
    input_node2 = ak.Input()
    output_node1 = ak.DenseBlock()(input_node1)
    output_node2 = ak.DenseBlock()(input_node2)
    output_node = ak.Merge()([output_node1, output_node2])
    head = ak.RegressionHead()
    output_node = head(output_node)
    head.outputs = output_node1

    with pytest.raises(ValueError) as info:
        graph_module.Graph(inputs=[input_node1, input_node2],
                           outputs=output_node)
    assert 'The network has a cycle.' in str(info.value)


def test_input_missing():
    input_node1 = ak.Input()
    input_node2 = ak.Input()
    output_node1 = ak.DenseBlock()(input_node1)
    output_node2 = ak.DenseBlock()(input_node2)
    output_node = ak.Merge()([output_node1, output_node2])
    output_node = ak.RegressionHead()(output_node)

    with pytest.raises(ValueError) as info:
        graph_module.Graph(inputs=input_node1,
                           outputs=output_node)
    assert 'A required input is missing for HyperModel' in str(info.value)


def test_graph_basics():
    input_node = ak.Input(shape=(30,))
    output_node = input_node
    output_node = ak.DenseBlock()(output_node)
    output_node = ak.RegressionHead(output_shape=(1,))(output_node)

    model = graph_module.Graph(
        inputs=input_node,
        outputs=output_node).build(kerastuner.HyperParameters())
    assert model.input_shape == (None, 30)
    assert model.output_shape == (None, 1)


def test_graph_save_load(tmp_path):
    input1 = ak.Input()
    input2 = ak.Input()
    output1 = ak.DenseBlock()(input1)
    output2 = ak.ConvBlock()(input2)
    output = ak.Merge()([output1, output2])
    output1 = ak.RegressionHead()(output)
    output2 = ak.ClassificationHead()(output)

    graph = graph_module.Graph(
        inputs=[input1, input2],
        outputs=[output1, output2],
        override_hps=[hp_module.Choice('dense_block_1/num_layers', [6], default=6)])
    config = graph.get_config()
    graph = graph_module.Graph.from_config(config)

    assert len(graph.inputs) == 2
    assert len(graph.outputs) == 2
    assert isinstance(graph.inputs[0].out_blocks[0], ak.DenseBlock)
    assert isinstance(graph.inputs[1].out_blocks[0], ak.ConvBlock)
    assert isinstance(graph.override_hps[0], hp_module.Choice)


def test_merge():
    input_node1 = ak.Input(shape=(30,))
    input_node2 = ak.Input(shape=(40,))
    output_node1 = ak.DenseBlock()(input_node1)
    output_node2 = ak.DenseBlock()(input_node2)
    output_node = ak.Merge()([output_node1, output_node2])
    output_node = ak.RegressionHead(output_shape=(1,))(output_node)

    model = graph_module.Graph(
        inputs=[input_node1, input_node2],
        outputs=output_node).build(kerastuner.HyperParameters())
    assert model.input_shape == [(None, 30), (None, 40)]
    assert model.output_shape == (None, 1)


def test_save_custom_metrics_loss(tmp_path):
    def custom_metric(y_pred, y_true):
        return 1

    def custom_loss(y_pred, y_true):
        return y_pred - y_true

    head = ak.ClassificationHead(
        loss=custom_loss, metrics=['accuracy', custom_metric])
    input_node = ak.Input()
    output_node = head(input_node)
    graph = graph_module.Graph(input_node, output_node)
    path = os.path.join(tmp_path, 'graph')
    graph.save(path)
    new_graph = graph_module.load_graph(
        path, custom_objects={
            'custom_metric': custom_metric,
            'custom_loss': custom_loss,
        })
    assert new_graph.blocks[0].metrics[1](0, 0) == 1
    assert new_graph.blocks[0].loss(3, 2) == 1



import kerastuner
import pytest
import tensorflow as tf
from tensorflow.python.util import nest

from autokeras import blocks
from tests import utils


def test_resnet_build_return_tensor():
    block = blocks.ResNetBlock()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 3), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_resnet_pretrained_build_return_tensor():
    block = blocks.ResNetBlock(pretrained=True)

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 3), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_resnet_pretrained_with_one_channel_input():
    block = blocks.ResNetBlock(pretrained=True)

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(28, 28, 1), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_resnet_pretrained_error_with_two_channels():
    block = blocks.ResNetBlock(pretrained=True)

    with pytest.raises(ValueError) as info:
        block.build(kerastuner.HyperParameters(),
                    tf.keras.Input(shape=(224, 224, 2), dtype=tf.float32))

    assert 'When pretrained is set to True' in str(info.value)


def test_resnet_deserialize_to_resnet():
    serialized_block = blocks.serialize(blocks.ResNetBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.ResNetBlock)


def test_resnet_get_config_has_all_attributes():
    block = blocks.ResNetBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.ResNetBlock.__init__).issubset(config.keys())


def test_resnet_wrong_version_error():
    with pytest.raises(ValueError) as info:
        blocks.ResNetBlock(version='abc')

    assert 'Expect version to be' in str(info.value)


def test_xception_build_return_tensor():
    block = blocks.XceptionBlock()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 2), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_xception_pretrained_build_return_tensor():
    block = blocks.XceptionBlock(pretrained=True)

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 3), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_xception_pretrained_with_one_channel_input():
    block = blocks.XceptionBlock(pretrained=True)

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(224, 224, 1), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_xception_pretrained_error_with_two_channels():
    block = blocks.XceptionBlock(pretrained=True)

    with pytest.raises(ValueError) as info:
        block.build(kerastuner.HyperParameters(),
                    tf.keras.Input(shape=(224, 224, 2), dtype=tf.float32))

    assert 'When pretrained is set to True' in str(info.value)


def test_xception_deserialize_to_xception():
    serialized_block = blocks.serialize(blocks.XceptionBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.XceptionBlock)


def test_xception_get_config_has_all_attributes():
    block = blocks.XceptionBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.XceptionBlock.__init__).issubset(config.keys())


def test_conv_build_return_tensor():
    block = blocks.ConvBlock()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 3), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_conv_deserialize_to_conv():
    serialized_block = blocks.serialize(blocks.ConvBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.ConvBlock)


def test_conv_get_config_has_all_attributes():
    block = blocks.ConvBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.ConvBlock.__init__).issubset(config.keys())


def test_rnn_build_return_tensor():
    block = blocks.RNNBlock()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 10), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_rnn_deserialize_to_rnn():
    serialized_block = blocks.serialize(blocks.RNNBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.RNNBlock)


def test_rnn_get_config_has_all_attributes():
    block = blocks.RNNBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.RNNBlock.__init__).issubset(config.keys())


def test_dense_build_return_tensor():
    block = blocks.DenseBlock()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32,), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_dense_deserialize_to_dense():
    serialized_block = blocks.serialize(blocks.DenseBlock())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.DenseBlock)


def test_dense_get_config_has_all_attributes():
    block = blocks.DenseBlock()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.DenseBlock.__init__).issubset(config.keys())


def test_embed_build_return_tensor():
    block = blocks.Embedding()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32,), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_embed_deserialize_to_embed():
    serialized_block = blocks.serialize(blocks.Embedding())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.Embedding)


def test_embed_get_config_has_all_attributes():
    block = blocks.Embedding()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.Embedding.__init__).issubset(config.keys())


def test_transformer_build_return_tensor():
    block = blocks.Transformer()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(64,), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_transformer_deserialize_to_transformer():
    serialized_block = blocks.serialize(blocks.Transformer())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.Transformer)


def test_transformer_get_config_has_all_attributes():
    block = blocks.Transformer()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.Transformer.__init__).issubset(config.keys())



from unittest import mock

import numpy as np

import autokeras as ak
from tests import utils


@mock.patch('autokeras.AutoModel.fit')
def test_txt_clf_fit_call_auto_model_fit(fit, tmp_path):
    auto_model = ak.TextClassifier(directory=tmp_path, seed=utils.SEED)

    auto_model.fit(
        x=np.array(['a b c', 'b b c']),
        y=np.array([1, 2]))

    assert fit.is_called


@mock.patch('autokeras.AutoModel.fit')
def test_txt_reg_fit_call_auto_model_fit(fit, tmp_path):
    auto_model = ak.TextRegressor(directory=tmp_path, seed=utils.SEED)

    auto_model.fit(
        x=np.array(['a b c', 'b b c']),
        y=np.array([1.0, 2.0]))

    assert fit.is_called



from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import pandas as pd

from autokeras import auto_model
from autokeras import blocks
from autokeras import nodes as input_module
from autokeras.engine import tuner
from autokeras.tasks.structured_data_mixin import StructuredDataMixin
from autokeras.tuners import greedy
from autokeras.utils import types


class SupervisedTimeseriesDataPipeline(StructuredDataMixin, auto_model.AutoModel):

    def __init__(self,
                 outputs,
                 column_names=None,
                 column_types=None,
                 lookback=None,
                 predict_from=1,
                 predict_until=None,
                 **kwargs):
        inputs = input_module.TimeseriesInput(lookback=lookback,
                                              column_names=column_names,
                                              column_types=column_types)
        self.check(column_names, column_types)
        super().__init__(inputs=inputs,
                         outputs=outputs,
                         **kwargs)
        self.predict_from = predict_from
        self.predict_until = predict_until
        self._target_col_name = None
        self.train_len = 0

    @staticmethod
    def _read_from_csv(x, y):
        df = pd.read_csv(x)
        target = df.pop(y).dropna().to_numpy()
        return df, target

    def fit(self,
            x=None,
            y=None,
            epochs=None,
            callbacks=None,
            validation_split=0.2,
            validation_data=None,
            **kwargs):
        # x is file path of training data
        if isinstance(x, str):
            self._target_col_name = y
            x, y = self._read_from_csv(x, y)
        if validation_data:
            x_val, y_val = validation_data
            if isinstance(x_val, str):
                validation_data = self._read_from_csv(x_val, y_val)

        self.train_len = len(y)

        super().fit(x=x[:self.train_len],
                    y=y[self.lookback-1:],
                    epochs=epochs,
                    callbacks=callbacks,
                    validation_split=validation_split,
                    validation_data=validation_data,
                    **kwargs)

    def predict(self, x, batch_size=32, **kwargs):
        x = self.read_for_predict(x)
        y_pred = super().predict(x=x,
                                 batch_size=batch_size,
                                 **kwargs)
        lower_bound = self.train_len + self.predict_from
        if self.predict_until is None:
            self.predict_until = len(y_pred)
        upper_bound = min(self.train_len + self.predict_until + 1, len(y_pred))
        return y_pred[lower_bound:upper_bound]

    def evaluate(self, x, y=None, batch_size=32, **kwargs):
        """Evaluate the best model for the given data.

        # Arguments
            x: String, numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                Testing data x. If the data is from a csv file, it should be a
                string specifying the path of the csv file of the testing data.
            y: String, numpy.ndarray, or tensorflow.Dataset. Testing data y.
                If the data is from a csv file, it should be a string corresponding
                to the label column.
            batch_size: Int. Defaults to 32.
            **kwargs: Any arguments supported by keras.Model.evaluate.

        # Returns
            Scalar test loss (if the model has a single output and no metrics) or
            list of scalars (if the model has multiple outputs and/or metrics).
            The attribute model.metrics_names will give you the display labels for
            the scalar outputs.
        """
        if isinstance(x, str):
            x, y = self._read_from_csv(x, y)
        return super().evaluate(x=x[:len(y)],
                                y=y[self.lookback-1:],
                                batch_size=batch_size,
                                **kwargs)


class TimeseriesForecaster(SupervisedTimeseriesDataPipeline):
    """AutoKeras time series data forecast class.

    # Arguments
        column_names: A list of strings specifying the names of the columns. The
            length of the list should be equal to the number of columns of the data.
            Defaults to None. If None, it will be obtained from the header of the csv
            file or the pandas.DataFrame.
        column_types: Dict. The keys are the column names. The values should either
            be 'numerical' or 'categorical', indicating the type of that column.
            Defaults to None. If not None, the column_names need to be specified.
            If None, it will be inferred from the data.
        lookback: Int. The range of history steps to consider for each prediction.
            For example, if lookback=n, the data in the range of [i - n, i - 1]
            is used to predict the value of step i. If unspecified, it will be tuned
            automatically.
        predict_from: Int. The starting point of the forecast for each sample (in
            number of steps) after the last time step in the input. If N is the last
            step in the input, then the first step of the predicted output will be
            N + predict_from. Defaults to 1 (which corresponds to starting the
            forecast immediately after the last step in the input).
        predict_until: Int. The end point of the forecast for each sample (in number
            of steps) after the last time step in the input. If N is the last step in
            the input, then the last step of the predicted output will be
            N + predict_until. If unspecified, it will predict till end of dataset.
            Defaults to None.
        loss: A Keras loss function. Defaults to use 'mean_squared_error'.
        metrics: A list of Keras metrics. Defaults to use 'mean_squared_error'.
        project_name: String. The name of the AutoModel. Defaults to
            'time_series_forecaster'.
        max_trials: Int. The maximum number of different Keras Models to try.
            The search may finish before reaching the max_trials. Defaults to 100.
        directory: String. The path to a directory for storing the search outputs.
            Defaults to None, which would create a folder with the name of the
            AutoModel in the current directory.
        objective: String. Name of model metric to minimize
            or maximize, e.g. 'val_accuracy'. Defaults to 'val_loss'.
        tuner: String or subclass of AutoTuner. If string, it should be one of
            'greedy', 'bayesian', 'hyperband' or 'random'. It can also be a subclass
            of AutoTuner. If left unspecified, it uses a task specific tuner, which
            first evaluates the most commonly used models for the task before
            exploring other models.
        overwrite: Boolean. Defaults to `False`. If `False`, reloads an existing
            project of the same name if one is found. Otherwise, overwrites the
            project.
        seed: Int. Random seed.
        **kwargs: Any arguments supported by AutoModel.
    """

    def __init__(self,
                 output_dim=None,
                 column_names: Optional[List[str]] = None,
                 column_types: Optional[Dict[str, str]] = None,
                 lookback: Optional[int] = None,
                 predict_from: int = 1,
                 predict_until: Optional[int] = None,
                 loss: types.LossType = 'mean_squared_error',
                 metrics: Optional[types.MetricsType] = None,
                 project_name: str = 'time_series_forecaster',
                 max_trials: int = 100,
                 directory: Union[str, Path, None] = None,
                 objective: str = 'val_loss',
                 tuner: Union[str, Type[tuner.AutoTuner]] = None,
                 overwrite: bool = False,
                 seed: Optional[int] = None,
                 **kwargs):
        if tuner is None:
            tuner = greedy.Greedy
        super().__init__(outputs=blocks.RegressionHead(output_dim=output_dim,
                                                       loss=loss,
                                                       metrics=metrics),
                         column_names=column_names,
                         column_types=column_types,
                         lookback=lookback,
                         predict_from=predict_from,
                         predict_until=predict_until,
                         project_name=project_name,
                         max_trials=max_trials,
                         directory=directory,
                         objective=objective,
                         tuner=tuner,
                         overwrite=overwrite,
                         seed=seed,
                         **kwargs)
        self.lookback = lookback
        self.predict_from = predict_from
        self.predict_until = predict_until

    def fit(self,
            x=None,
            y=None,
            validation_split=0.2,
            validation_data=None,
            **kwargs):
        """Search for the best model and hyperparameters for the task.

        # Arguments
            x: String, numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                Training data x. If the data is from a csv file, it should be a
                string specifying the path of the csv file of the training data.
            y: String, a list of string(s), numpy.ndarray, pandas.DataFrame or
                tensorflow.Dataset. Training data y.
                If the data is from a csv file, it should be a list of string(s)
                specifying the name(s) of the column(s) need to be forecasted.
                If it is multivariate forecasting, y should be a list of more than
                one column names. If it is univariate forecasting, y should be a
                string or a list of one string.
            validation_split: Float between 0 and 1. Defaults to 0.2.
                Fraction of the training data to be used as validation data.
                The model will set apart this fraction of the training data,
                will not train on it, and will evaluate
                the loss and any model metrics
                on this data at the end of each epoch.
                The validation data is selected from the last samples
                in the `x` and `y` data provided, before shuffling. This argument is
                not supported when `x` is a dataset.
                The best model found would be fit on the entire dataset including the
                validation data.
            validation_data: Data on which to evaluate the loss and any model metrics
                at the end of each epoch. The model will not be trained on this data.
                `validation_data` will override `validation_split`. The type of the
                validation data should be the same as the training data.
                The best model found would be fit on the training dataset without the
                validation data.
            **kwargs: Any arguments supported by keras.Model.fit.
        """
        super().fit(x=x,
                    y=y,
                    validation_split=validation_split,
                    validation_data=validation_data,
                    **kwargs)

    def predict(self, x=None, batch_size=32, **kwargs):
        """Predict the output for a given testing data.

        # Arguments
            x: String, numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                Testing data x. If the data is from a csv file, it should be a
                string specifying the path of the csv file of the testing data.
            batch_size: Int. Defaults to 32.
            **kwargs: Any arguments supported by keras.Model.predict.

        # Returns
            A list of numpy.ndarray objects or a single numpy.ndarray.
            The predicted results.
        """
        return super().predict(x=x, batch_size=batch_size, **kwargs)

    def fit_and_predict(self,
                        x=None,
                        y=None,
                        validation_split=0.2,
                        validation_data=None,
                        batch_size=32,
                        **kwargs):
        """Search for the best model and then predict for remaining data points.

        # Arguments
            x: String, numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                Training data x. If the data is from a csv file, it should be a
                string specifying the path of the csv file of the training data.
            y: String, a list of string(s), numpy.ndarray, pandas.DataFrame or
                tensorflow.Dataset. Training data y.
                If the data is from a csv file, it should be a list of string(s)
                specifying the name(s) of the column(s) need to be forecasted.
                If it is multivariate forecasting, y should be a list of more than
                one column names. If it is univariate forecasting, y should be a
                string or a list of one string.
            validation_split: Float between 0 and 1. Defaults to 0.2.
                Fraction of the training data to be used as validation data.
                The model will set apart this fraction of the training data,
                will not train on it, and will evaluate
                the loss and any model metrics
                on this data at the end of each epoch.
                The validation data is selected from the last samples
                in the `x` and `y` data provided, before shuffling. This argument is
                not supported when `x` is a dataset.
                The best model found would be fit on the entire dataset including the
                validation data.
            validation_data: Data on which to evaluate the loss and any model metrics
                at the end of each epoch. The model will not be trained on this data.
                `validation_data` will override `validation_split`. The type of the
                validation data should be the same as the training data.
                The best model found would be fit on the training dataset without the
                validation data.
            batch_size: Int. Defaults to 32.
            **kwargs: Any arguments supported by keras.Model.fit.
        """
        self.fit(x=x,
                 y=y,
                 validation_split=validation_split,
                 validation_data=validation_data,
                 **kwargs)

        return self.predict(x=x, batch_size=batch_size)


class TimeseriesClassifier(SupervisedTimeseriesDataPipeline):
    """"AutoKeras time series data classification class.

    # Arguments
        column_names: A list of strings specifying the names of the columns. The
            length of the list should be equal to the number of columns of the data.
            Defaults to None. If None, it will be obtained from the header of the csv
            file or the pandas.DataFrame.
        column_types: Dict. The keys are the column names. The values should either
            be 'numerical' or 'categorical', indicating the type of that column.
            Defaults to None. If not None, the column_names need to be specified.
            If None, it will be inferred from the data.
        lookback: Int. The range of history steps to consider for each prediction.
            For example, if lookback=n, the data in the range of [i - n, i - 1]
            is used to predict the value of step i. If unspecified, it will be tuned
            automatically.
        predict_from: Int. The starting point of the forecast for each sample (in
            number of steps) after the last time step in the input. If N is the last
            step in the input, then the first step of the predicted output will be
            N + predict_from. Defaults to 1 (which corresponds to starting the
            forecast immediately after the last step in the input).
        predict_until: Int. The end point of the forecast for each sample (in number
            of steps) after the last time step in the input. If N is the last step in
            the input, then the last step of the predicted output will be
            N + predict_until. If unspecified, it will predict till end of dataset.
            Defaults to None.
        loss: A Keras loss function. Defaults to use 'mean_squared_error'.
        metrics: A list of Keras metrics. Defaults to use 'mean_squared_error'.
        project_name: String. The name of the AutoModel. Defaults to
            'time_series_forecaster'.
        max_trials: Int. The maximum number of different Keras Models to try.
            The search may finish before reaching the max_trials. Defaults to 100.
        directory: String. The path to a directory for storing the search outputs.
            Defaults to None, which would create a folder with the name of the
            AutoModel in the current directory.
        objective: String. Name of model metric to minimize
            or maximize, e.g. 'val_accuracy'. Defaults to 'val_loss'.
        overwrite: Boolean. Defaults to `False`. If `False`, reloads an existing
            project of the same name if one is found. Otherwise, overwrites the
            project.
        seed: Int. Random seed.
        **kwargs: Any arguments supported by AutoModel.
    """

    def __init__(self,
                 output_dim=None,
                 column_names=None,
                 column_types=None,
                 lookback=None,
                 predict_from=1,
                 predict_until=None,
                 loss='mean_squared_error',
                 metrics=None,
                 project_name='time_series_classifier',
                 max_trials=100,
                 directory=None,
                 objective='val_loss',
                 overwrite=False,
                 seed=None,
                 **kwargs):
        raise NotImplementedError

    def fit(self,
            x=None,
            y=None,
            validation_split=0.2,
            validation_data=None,
            **kwargs):
        """Search for the best model and hyperparameters for the task.

        # Arguments
            x: String, numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                Training data x. If the data is from a csv file, it should be a
                string specifying the path of the csv file of the training data.
            y: String, a list of string(s), numpy.ndarray, pandas.DataFrame or
                tensorflow.Dataset. Training data y.
                If the data is from a csv file, it should be a list of string(s)
                specifying the name(s) of the column(s) need to be forecasted.
                If it is multivariate forecasting, y should be a list of more than
                one column names. If it is univariate forecasting, y should be a
                string or a list of one string.
            validation_split: Float between 0 and 1. Defaults to 0.2.
                Fraction of the training data to be used as validation data.
                The model will set apart this fraction of the training data,
                will not train on it, and will evaluate
                the loss and any model metrics
                on this data at the end of each epoch.
                The validation data is selected from the last samples
                in the `x` and `y` data provided, before shuffling. This argument is
                not supported when `x` is a dataset.
                The best model found would be fit on the entire dataset including the
                validation data.
            validation_data: Data on which to evaluate the loss and any model metrics
                at the end of each epoch. The model will not be trained on this data.
                `validation_data` will override `validation_split`. The type of the
                validation data should be the same as the training data.
                The best model found would be fit on the training dataset without the
                validation data.
            **kwargs: Any arguments supported by keras.Model.fit.
        """
        raise NotImplementedError

    def predict(self, x=None, batch_size=32, **kwargs):
        """Predict the output for a given testing data.

        # Arguments
            x: String, numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                Testing data x, it should also contain the training data used as,
                subsequent predictions depend on them. If the data is from a csv
                file, it should be a string specifying the path of the csv file
                of the testing data.
            batch_size: Int. Defaults to 32.
            **kwargs: Any arguments supported by keras.Model.predict.

        # Returns
            A list of numpy.ndarray objects or a single numpy.ndarray.
            The predicted results.
        """
        raise NotImplementedError

    def fit_and_predict(self,
                        x=None,
                        y=None,
                        validation_split=0.2,
                        validation_data=None,
                        batch_size=32,
                        **kwargs):
        """Search for the best model and then predict for remaining data points.

        # Arguments
            x: String, numpy.ndarray, pandas.DataFrame or tensorflow.Dataset.
                Training and Test data x. If the data is from a csv file, it
                should be a string specifying the path of the csv file of the
                training data.
            y: String, a list of string(s), numpy.ndarray, pandas.DataFrame or
                tensorflow.Dataset. Training data y.
                If the data is from a csv file, it should be a list of string(s)
                specifying the name(s) of the column(s) need to be forecasted.
                If it is multivariate forecasting, y should be a list of more than
                one column names. If it is univariate forecasting, y should be a
                string or a list of one string.
            validation_split: Float between 0 and 1. Defaults to 0.2.
                Fraction of the training data to be used as validation data.
                The model will set apart this fraction of the training data,
                will not train on it, and will evaluate
                the loss and any model metrics
                on this data at the end of each epoch.
                The validation data is selected from the last samples
                in the `x` and `y` data provided, before shuffling. This argument is
                not supported when `x` is a dataset.
                The best model found would be fit on the entire dataset including the
                validation data.
            validation_data: Data on which to evaluate the loss and any model metrics
                at the end of each epoch. The model will not be trained on this data.
                `validation_data` will override `validation_split`. The type of the
                validation data should be the same as the training data.
                The best model found would be fit on the training dataset without the
                validation data.
            batch_size: Int. Defaults to 32.
            **kwargs: Any arguments supported by keras.Model.fit.
        """
        raise NotImplementedError



from unittest import mock

import numpy as np
import pytest
import tensorflow as tf

import autokeras as ak
from tests import utils


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_evaluate(tuner_fn, tmp_path):
    x_train = np.random.rand(100, 32)
    y_train = np.random.rand(100, 1)

    input_node = ak.Input()
    output_node = input_node
    output_node = ak.DenseBlock()(output_node)
    output_node = ak.RegressionHead()(output_node)

    auto_model = ak.AutoModel(input_node,
                              output_node,
                              directory=tmp_path,
                              max_trials=1)
    auto_model.fit(x_train, y_train, epochs=1, validation_data=(x_train, y_train))
    auto_model.evaluate(x_train, y_train)
    assert tuner_fn.called


def get_single_io_auto_model(tmp_path):
    return ak.AutoModel(ak.ImageInput(),
                        ak.RegressionHead(),
                        directory=tmp_path,
                        max_trials=2)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_auto_model_predict(tuner_fn, tmp_path):
    x_train = np.random.rand(100, 32, 32, 3)
    y_train = np.random.rand(100, 1)

    auto_model = get_single_io_auto_model(tmp_path)
    auto_model.fit(x_train, y_train, epochs=2, validation_split=0.2)
    auto_model.predict(x_train)
    assert tuner_fn.called


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_final_fit_concat(tuner_fn, tmp_path):
    tuner = tuner_fn.return_value.return_value

    x_train = np.random.rand(100, 32, 32, 3)
    y_train = np.random.rand(100, 1)

    auto_model = get_single_io_auto_model(tmp_path)
    auto_model.fit(x_train, y_train, epochs=2, validation_split=0.2)
    assert auto_model._split_dataset
    assert tuner.search.call_args_list[0][1]['fit_on_val_data']


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_final_fit_not_concat(tuner_fn, tmp_path):
    tuner = tuner_fn.return_value.return_value

    x_train = np.random.rand(100, 32, 32, 3)
    y_train = np.random.rand(100, 1)

    auto_model = get_single_io_auto_model(tmp_path)
    auto_model.fit(x_train, y_train, epochs=2, validation_data=(x_train, y_train))
    assert not auto_model._split_dataset
    assert not tuner.search.call_args_list[0][1]['fit_on_val_data']


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_overwrite(tuner_fn, tmp_path):
    tuner_class = tuner_fn.return_value

    x_train = np.random.rand(100, 32, 32, 3)
    y_train = np.random.rand(100, 1)

    auto_model = get_single_io_auto_model(tmp_path)
    auto_model.fit(x_train, y_train, epochs=2, validation_data=(x_train, y_train))
    assert not tuner_class.call_args_list[0][1]['overwrite']


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_export_model(tuner_fn, tmp_path):
    tuner_class = tuner_fn.return_value
    tuner = tuner_class.return_value

    x_train = np.random.rand(100, 32, 32, 3)
    y_train = np.random.rand(100, 1)

    auto_model = get_single_io_auto_model(tmp_path)
    auto_model.fit(x_train, y_train, epochs=2, validation_data=(x_train, y_train))
    auto_model.export_model()
    assert tuner.get_best_model.called


def get_multi_io_auto_model(tmp_path):
    return ak.AutoModel([ak.ImageInput(), ak.ImageInput()],
                        [ak.RegressionHead(), ak.RegressionHead()],
                        directory=tmp_path,
                        max_trials=2,
                        overwrite=False)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_multi_io_with_tf_dataset(tuner_fn, tmp_path):
    auto_model = get_multi_io_auto_model(tmp_path)
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices(((x1, x1), (y1, y1)))
    auto_model.fit(dataset, epochs=2)

    for adapter in auto_model._input_adapters + auto_model._output_adapters:
        assert adapter.shape is not None


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_single_nested_dataset(tuner_fn, tmp_path):
    auto_model = ak.AutoModel(ak.ImageInput(),
                              ak.RegressionHead(),
                              directory=tmp_path,
                              max_trials=2,
                              overwrite=False)
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices(((x1,), y1))
    auto_model.fit(dataset, epochs=2)

    for adapter in auto_model._input_adapters + auto_model._output_adapters:
        assert adapter.shape is not None


def dataset_error(x, y, validation_data, message, tmp_path):
    auto_model = get_multi_io_auto_model(tmp_path)
    with pytest.raises(ValueError) as info:
        auto_model.fit(x, y, epochs=2, validation_data=validation_data)
    assert message in str(info.value)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_data_io_consistency_input(tuner_fn, tmp_path):
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices(((x1,), (y1, y1)))
    dataset_error(dataset, None, dataset, 'Expect x to have', tmp_path)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_data_io_consistency_output(tuner_fn, tmp_path):
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices(((x1, x1), (y1,)))
    dataset_error(dataset, None, dataset, 'Expect y to have', tmp_path)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_data_io_consistency_validation(tuner_fn, tmp_path):
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices(((x1, x1), (y1, y1)))
    val_dataset = tf.data.Dataset.from_tensor_slices(((x1,), (y1, y1)))
    dataset_error(dataset, None, val_dataset,
                  'Expect x in validation_data to have', tmp_path)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_dataset_and_y(tuner_fn, tmp_path):
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    x = tf.data.Dataset.from_tensor_slices((x1, x1))
    y = tf.data.Dataset.from_tensor_slices((y1, y1))
    val_dataset = tf.data.Dataset.from_tensor_slices(((x1,), (y1, y1)))
    dataset_error(x, y, val_dataset,
                  'Expect y is None', tmp_path)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_multi_input_predict(tuner_fn, tmp_path):
    auto_model = get_multi_io_auto_model(tmp_path)
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices(((x1, x1), (y1, y1)))
    auto_model.fit(dataset, None, epochs=2, validation_data=dataset)

    dataset2 = tf.data.Dataset.from_tensor_slices(((x1, x1),))
    auto_model.predict(dataset2)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_multi_input_predict2(tuner_fn, tmp_path):
    auto_model = get_multi_io_auto_model(tmp_path)
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices(((x1, x1), (y1, y1)))
    auto_model.fit(dataset, None, epochs=2, validation_data=dataset)

    dataset2 = tf.data.Dataset.from_tensor_slices((x1, x1))
    auto_model.predict(dataset2)


@mock.patch('autokeras.auto_model.get_tuner_class')
def test_single_input_predict(tuner_fn, tmp_path):
    auto_model = get_single_io_auto_model(tmp_path)
    x1 = utils.generate_data()
    y1 = utils.generate_data(shape=(1,))
    dataset = tf.data.Dataset.from_tensor_slices((x1, y1))
    auto_model.fit(dataset, None, epochs=2, validation_data=dataset)

    dataset2 = tf.data.Dataset.from_tensor_slices((x1, y1))
    auto_model.predict(dataset2)



from unittest import mock

import autokeras as ak
from tests import utils


@mock.patch('autokeras.AutoModel.fit')
def test_img_clf_fit_call_auto_model_fit(fit, tmp_path):
    auto_model = ak.ImageClassifier(directory=tmp_path, seed=utils.SEED)

    auto_model.fit(
        x=utils.generate_data(num_instances=100, shape=(32, 32, 3)),
        y=utils.generate_one_hot_labels(num_instances=100, num_classes=10))

    assert fit.is_called


@mock.patch('autokeras.AutoModel.fit')
def test_img_reg_fit_call_auto_model_fit(fit, tmp_path):
    auto_model = ak.ImageRegressor(directory=tmp_path, seed=utils.SEED)

    auto_model.fit(
        x=utils.generate_data(num_instances=100, shape=(32, 32, 3)),
        y=utils.generate_data(num_instances=100, shape=(1,)))

    assert fit.is_called


@mock.patch('autokeras.AutoModel.fit')
def test_img_seg_fit_call_auto_model_fit(fit, tmp_path):
    auto_model = ak.tasks.image.ImageSegmenter(
        directory=tmp_path, seed=utils.SEED)

    auto_model.fit(
        x=utils.generate_data(num_instances=100, shape=(32, 32, 3)),
        y=utils.generate_data(num_instances=100, shape=(32, 32)))

    assert fit.is_called



import kerastuner

from autokeras.engine import tuner as tuner_module


class RandomSearch(kerastuner.RandomSearch, tuner_module.AutoTuner):
    """KerasTuner RandomSearch with preprocessing layer tuning."""
    pass



import kerastuner
import tensorflow as tf
from tensorflow.python.util import nest

from autokeras import blocks
from tests import utils


def test_merge_build_return_tensor():
    block = blocks.Merge()

    outputs = block.build(
        kerastuner.HyperParameters(),
        [tf.keras.Input(shape=(32,), dtype=tf.float32),
         tf.keras.Input(shape=(4, 8), dtype=tf.float32)])

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_merge_deserialize_to_merge():
    serialized_block = blocks.serialize(blocks.Merge())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.Merge)


def test_merge_get_config_has_all_attributes():
    block = blocks.Merge()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.Merge.__init__).issubset(config.keys())


def test_temporal_build_return_tensor():
    block = blocks.TemporalReduction()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 10), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_temporal_deserialize_to_temporal():
    serialized_block = blocks.serialize(blocks.TemporalReduction())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.TemporalReduction)


def test_temporal_get_config_has_all_attributes():
    block = blocks.TemporalReduction()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.TemporalReduction.__init__).issubset(config.keys())


def test_spatial_build_return_tensor():
    block = blocks.SpatialReduction()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 3), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_spatial_deserialize_to_spatial():
    serialized_block = blocks.serialize(blocks.SpatialReduction())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.SpatialReduction)


def test_spatial_get_config_has_all_attributes():
    block = blocks.SpatialReduction()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.SpatialReduction.__init__).issubset(config.keys())



import kerastuner
import tensorflow as tf
from tensorflow.python.util import nest

from autokeras import blocks
from tests import utils


def test_augment_build_return_tensor():
    block = blocks.ImageAugmentation()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(32, 32, 3), dtype=tf.float32))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_augment_deserialize_to_augment():
    serialized_block = blocks.serialize(blocks.ImageAugmentation())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.ImageAugmentation)


def test_augment_get_config_has_all_attributes():
    block = blocks.ImageAugmentation()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.ImageAugmentation.__init__).issubset(config.keys())


def test_ngram_build_return_tensor():
    block = blocks.TextToNgramVector()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(1,), dtype=tf.string))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_ngram_deserialize_to_ngram():
    serialized_block = blocks.serialize(blocks.TextToNgramVector())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.TextToNgramVector)


def test_ngram_get_config_has_all_attributes():
    block = blocks.TextToNgramVector()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.TextToNgramVector.__init__).issubset(config.keys())


def test_int_seq_build_return_tensor():
    block = blocks.TextToIntSequence()

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(1,), dtype=tf.string))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_int_seq_deserialize_to_int_seq():
    serialized_block = blocks.serialize(blocks.TextToIntSequence())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.TextToIntSequence)


def test_int_seq_get_config_has_all_attributes():
    block = blocks.TextToIntSequence()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.TextToIntSequence.__init__).issubset(config.keys())


def test_cat_to_num_build_return_tensor():
    block = blocks.CategoricalToNumerical()
    block.column_names = ['a']
    block.column_types = {'a': 'num'}

    outputs = block.build(
        kerastuner.HyperParameters(),
        tf.keras.Input(shape=(1,), dtype=tf.string))

    assert len(nest.flatten(outputs)) == 1
    assert isinstance(nest.flatten(outputs)[0], tf.Tensor)


def test_cat_to_num_deserialize_to_cat_to_num():
    serialized_block = blocks.serialize(blocks.CategoricalToNumerical())

    block = blocks.deserialize(serialized_block)

    assert isinstance(block, blocks.CategoricalToNumerical)


def test_cat_to_num_get_config_has_all_attributes():
    block = blocks.CategoricalToNumerical()

    config = block.get_config()

    assert utils.get_func_args(
        blocks.CategoricalToNumerical.__init__).issubset(config.keys())



from autokeras.tasks.image import ImageClassifier
from autokeras.tasks.image import ImageRegressor
from autokeras.tasks.structured_data import StructuredDataClassifier
from autokeras.tasks.structured_data import StructuredDataRegressor
from autokeras.tasks.text import TextClassifier
from autokeras.tasks.text import TextRegressor
from autokeras.tasks.time_series_forecaster import TimeseriesForecaster



import kerastuner
import tensorflow as tf
from tensorflow.python.util import nest

from autokeras.engine import node as node_module
from autokeras.engine import serializable
from autokeras.utils import utils


class Block(kerastuner.HyperModel, serializable.Serializable):
    """The base class for different Block.

    The Block can be connected together to build the search space
    for an AutoModel. Notably, many args in the __init__ function are defaults to
    be a tunable variable when not specified by the user.

    # Arguments
        name: String. The name of the block. If unspecified, it will be set
            automatically with the class name.
    """

    def __init__(self, name: str = None, **kwargs):
        if not name:
            prefix = self.__class__.__name__
            name = prefix + '_' + str(tf.keras.backend.get_uid(prefix))
            name = utils.to_snake_case(name)
        super().__init__(name=name, **kwargs)
        self.inputs = None
        self.outputs = None
        self._num_output_node = 1

    def _build_wrapper(self, hp, *args, **kwargs):
        with hp.name_scope(self.name):
            return super()._build_wrapper(hp, *args, **kwargs)

    def __call__(self, inputs):
        """Functional API.

        # Arguments
            inputs: A list of input node(s) or a single input node for the block.

        # Returns
            list: A list of output node(s) of the Block.
        """
        self.inputs = nest.flatten(inputs)
        for input_node in self.inputs:
            if not isinstance(input_node, node_module.Node):
                raise TypeError('Expect the inputs to block {name} to be '
                                'a Node, but got {type}.'.format(
                                    name=self.name,
                                    type=type(input_node)))
            input_node.add_out_block(self)
        self.outputs = []
        for _ in range(self._num_output_node):
            output_node = node_module.Node()
            output_node.add_in_block(self)
            self.outputs.append(output_node)
        return self.outputs

    def build(self, hp, inputs=None):
        """Build the Block into a real Keras Model.

        The subclasses should override this function and return the output node.

        # Arguments
            hp: HyperParameters. The hyperparameters for building the model.
            inputs: A list of input node(s).
        """
        return super().build(hp)

    def get_config(self):
        """Get the configuration of the preprocessor.

        # Returns
            A dictionary of configurations of the preprocessor.
        """
        return {'name': self.name, 'tunable': self.tunable}



import copy

import kerastuner
import pytest
import tensorflow as tf

import autokeras as ak
from autokeras.tuners import task_specific


@pytest.fixture
def clear_session():
    yield
    tf.keras.backend.clear_session()


def test_img_clf_init_hp0_equals_hp_of_a_model(clear_session, tmp_path):
    clf = ak.ImageClassifier(directory=tmp_path)
    clf.inputs[0].shape = (32, 32, 3)
    clf.outputs[0].in_blocks[0].output_shape = (10,)
    init_hp = task_specific.IMAGE_CLASSIFIER[0]
    hp = kerastuner.HyperParameters()
    hp.values = copy.copy(init_hp)

    clf.tuner.hypermodel.build(hp)

    assert set(init_hp.keys()) == set(hp._hps.keys())


def test_img_clf_init_hp1_equals_hp_of_a_model(clear_session, tmp_path):
    clf = ak.ImageClassifier(directory=tmp_path)
    clf.inputs[0].shape = (32, 32, 3)
    clf.outputs[0].in_blocks[0].output_shape = (10,)
    init_hp = task_specific.IMAGE_CLASSIFIER[1]
    hp = kerastuner.HyperParameters()
    hp.values = copy.copy(init_hp)

    clf.tuner.hypermodel.build(hp)

    assert set(init_hp.keys()) == set(hp._hps.keys())


def test_img_clf_init_hp2_equals_hp_of_a_model(clear_session, tmp_path):
    clf = ak.ImageClassifier(directory=tmp_path)
    clf.inputs[0].shape = (32, 32, 3)
    clf.outputs[0].in_blocks[0].output_shape = (10,)
    init_hp = task_specific.IMAGE_CLASSIFIER[2]
    hp = kerastuner.HyperParameters()
    hp.values = copy.copy(init_hp)

    clf.tuner.hypermodel.build(hp)

    assert set(init_hp.keys()) == set(hp._hps.keys())


def test_txt_clf_init_hp0_equals_hp_of_a_model(clear_session, tmp_path):
    clf = ak.TextClassifier(directory=tmp_path)
    clf.inputs[0].shape = (1,)
    clf.outputs[0].in_blocks[0].output_shape = (10,)
    init_hp = task_specific.TEXT_CLASSIFIER[0]
    hp = kerastuner.HyperParameters()
    hp.values = copy.copy(init_hp)

    clf.tuner.hypermodel.build(hp)

    assert set(init_hp.keys()) == set(hp._hps.keys())


def test_txt_clf_init_hp1_equals_hp_of_a_model(clear_session, tmp_path):
    clf = ak.TextClassifier(directory=tmp_path)
    clf.inputs[0].shape = (1,)
    clf.outputs[0].in_blocks[0].output_shape = (10,)
    init_hp = task_specific.TEXT_CLASSIFIER[1]
    hp = kerastuner.HyperParameters()
    hp.values = copy.copy(init_hp)

    clf.tuner.hypermodel.build(hp)

    assert set(init_hp.keys()) == set(hp._hps.keys())



import numpy as np
import tensorflow as tf
from tensorflow.python.util import nest


def batched(dataset):
    shape = nest.flatten(dataset_shape(dataset))[0]
    return len(shape) > 0 and shape[0] is None


def batch_dataset(dataset, batch_size):
    if batched(dataset):
        return dataset
    return dataset.batch(batch_size)


def split_dataset(dataset, validation_split):
    """Split dataset into training and validation.

    # Arguments
        dataset: tf.data.Dataset. The entire dataset to be split.
        validation_split: Float. The split ratio for the validation set.

    # Raises
        ValueError: If the dataset provided is too small to be split.

    # Returns
        A tuple of two tf.data.Dataset. The training set and the validation set.
    """
    num_instances = dataset.reduce(np.int64(0), lambda x, _: x + 1).numpy()
    if num_instances < 2:
        raise ValueError('The dataset should at least contain 2 '
                         'batches to be split.')
    validation_set_size = min(
        max(int(num_instances * validation_split), 1),
        num_instances - 1)
    train_set_size = num_instances - validation_set_size
    train_dataset = dataset.take(train_set_size)
    validation_dataset = dataset.skip(train_set_size)
    return train_dataset, validation_dataset


def dataset_shape(dataset):
    return tf.compat.v1.data.get_output_shapes(dataset)


def is_label(y):
    """Check if the targets are one-hot encoded or plain labels.

    # Arguments
        y: numpy.ndarray. The targets.

    # Returns
        Boolean. Whether the targets are plain label, not encoded.
    """
    return len(y.flatten()) == len(y)



import kerastuner

from autokeras.engine import serializable


class Preprocessor(kerastuner.HyperModel, serializable.Serializable):
    """Input data preprocessor search space.

    This class defines the search space for input data preprocessor. A
    preprocessor transforms the dataset using `tf.data` operations.
    """

    def build(self, hp, x):
        """Build the `tf.data` input preprocessor.

        # Arguments
            hp: `HyperParameters` instance. The hyperparameters for building the
                model.
            x: `tf.data.Dataset` instance. The input data for preprocessing.

        # Returns
            `tf.data.Dataset`. The preprocessed data to pass to the model.
        """
        raise NotImplementedError



from autokeras.tuners import greedy

IMAGE_CLASSIFIER = [{
    'image_block_1/block_type': 'vanilla',
    'image_block_1/normalize': True,
    'image_block_1/augment': False,
    'image_block_1/conv_block_1/kernel_size': 3,
    'image_block_1/conv_block_1/num_blocks': 1,
    'image_block_1/conv_block_1/num_layers': 2,
    'image_block_1/conv_block_1/max_pooling': True,
    'image_block_1/conv_block_1/separable': False,
    'image_block_1/conv_block_1/dropout': 0.25,
    'image_block_1/conv_block_1/filters_0_0': 32,
    'image_block_1/conv_block_1/filters_0_1': 64,
    'classification_head_1/spatial_reduction_1/reduction_type': 'flatten',
    'classification_head_1/dropout': 0.5,
    'optimizer': 'adam',
    'learning_rate': 1e-3,
}, {
    'image_block_1/block_type': 'resnet',
    'image_block_1/normalize': True,
    'image_block_1/augment': True,
    'image_block_1/image_augmentation_1/horizontal_flip': True,
    'image_block_1/image_augmentation_1/vertical_flip': False,
    'image_block_1/image_augmentation_1/contrast_factor': 0.0,
    'image_block_1/image_augmentation_1/rotation_factor': 0.0,
    'image_block_1/image_augmentation_1/translation_factor': 0.1,
    'image_block_1/image_augmentation_1/zoom_factor': 0.0,
    'image_block_1/res_net_block_1/pretrained': True,
    'image_block_1/res_net_block_1/version': 'resnet50',
    'image_block_1/res_net_block_1/trainable': True,
    'image_block_1/res_net_block_1/imagenet_size': True,
    'classification_head_1/spatial_reduction_1/reduction_type': 'global_avg',
    'classification_head_1/dropout': 0,
    'optimizer': 'adam',
    'learning_rate': 1e-5,
}, {
    'image_block_1/block_type': 'resnet',
    'image_block_1/normalize': True,
    'image_block_1/augment': True,
    'image_block_1/image_augmentation_1/horizontal_flip': True,
    'image_block_1/image_augmentation_1/vertical_flip': True,
    'image_block_1/image_augmentation_1/contrast_factor': 0.0,
    'image_block_1/image_augmentation_1/rotation_factor': 0.0,
    'image_block_1/image_augmentation_1/translation_factor': 0.1,
    'image_block_1/image_augmentation_1/zoom_factor': 0.0,
    'image_block_1/res_net_block_1/pretrained': False,
    'image_block_1/res_net_block_1/version': 'resnet50',
    'image_block_1/res_net_block_1/imagenet_size': False,
    'classification_head_1/spatial_reduction_1/reduction_type': 'global_avg',
    'classification_head_1/dropout': 0,
    'optimizer': 'adam',
    'learning_rate': 1e-3,
}]

TEXT_CLASSIFIER = [{
    'text_block_1/block_type': 'transformer',
    'classification_head_1/dropout': 0,
    'optimizer': 'adam',
    'learning_rate': 1e-3,
    'text_block_1/max_tokens': 20000,
    'text_block_1/text_to_int_sequence_1/output_sequence_length': 200,
    'text_block_1/transformer_1/pretraining': 'none',
    'text_block_1/transformer_1/embedding_dim': 32,
    'text_block_1/transformer_1/num_heads': 2,
    'text_block_1/transformer_1/dense_dim': 32,
    'text_block_1/transformer_1/dropout': 0.25,
    'text_block_1/spatial_reduction_1/reduction_type': 'global_avg',
    'text_block_1/dense_block_1/num_layers': 1,
    'text_block_1/dense_block_1/use_batchnorm': False,
    'text_block_1/dense_block_1/dropout': 0.5,
    'text_block_1/dense_block_1/units_0': 20,
}, {
    'text_block_1/block_type': 'vanilla',
    'classification_head_1/dropout': 0,
    'text_block_1/max_tokens': 5000,
    'text_block_1/conv_block_1/separable': False,
    'text_block_1/text_to_int_sequence_1/output_sequence_length': 512,
    'text_block_1/embedding_1/pretraining': 'none',
    'text_block_1/embedding_1/embedding_dim': 64,
    'text_block_1/embedding_1/dropout': 0.25,
    'text_block_1/conv_block_1/kernel_size': 5,
    'text_block_1/conv_block_1/num_blocks': 1,
    'text_block_1/conv_block_1/num_layers': 1,
    'text_block_1/conv_block_1/max_pooling': False,
    'text_block_1/conv_block_1/dropout': 0,
    'text_block_1/conv_block_1/filters_0_0': 256,
    'text_block_1/spatial_reduction_1/reduction_type': 'global_max',
    'text_block_1/dense_block_1/num_layers': 1,
    'text_block_1/dense_block_1/use_batchnorm': False,
    'text_block_1/dense_block_1/dropout': 0.5,
    'text_block_1/dense_block_1/units_0': 256,
    'optimizer': 'adam',
    'learning_rate': 1e-3,
}]


class ImageClassifierTuner(greedy.Greedy):
    def __init__(self, **kwargs):
        super().__init__(
            initial_hps=IMAGE_CLASSIFIER,
            **kwargs)


class TextClassifierTuner(greedy.Greedy):
    def __init__(self, **kwargs):
        super().__init__(
            initial_hps=TEXT_CLASSIFIER,
            **kwargs)



import numpy as np
import pandas as pd
import tensorflow as tf

from autokeras import encoders
from autokeras.engine import adapter as adapter_module
from autokeras.utils import data_utils


class HeadAdapter(adapter_module.Adapter):

    def __init__(self, name, **kwargs):
        super().__init__(**kwargs)
        self.name = name

    def check(self, dataset):
        supported_types = (tf.data.Dataset, np.ndarray, pd.DataFrame, pd.Series)
        if not isinstance(dataset, supported_types):
            raise TypeError('Expect the target data of {name} to be tf.data.Dataset,'
                            ' np.ndarray, pd.DataFrame or pd.Series, but got {type}.'
                            .format(name=self.name, type=type(dataset)))

    def convert_to_dataset(self, dataset):
        if isinstance(dataset, np.ndarray):
            if len(dataset.shape) == 1:
                dataset = dataset.reshape(-1, 1)
        if isinstance(dataset, pd.DataFrame):
            dataset = dataset.values
        if isinstance(dataset, pd.Series):
            dataset = dataset.values.reshape(-1, 1)
        return super().convert_to_dataset(dataset)

    def postprocess(self, y):
        """Postprocess the output of the Keras Model."""
        return y

    def get_config(self):
        config = super().get_config()
        config.update({
            'name': self.name,
        })
        return config


class ClassificationHeadAdapter(HeadAdapter):

    def __init__(self,
                 num_classes=None,
                 multi_label=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.num_classes = num_classes
        self.label_encoder = None
        self.multi_label = multi_label

    def get_config(self):
        config = super().get_config()
        config.update({
            'encoder': encoders.serialize(self.label_encoder),
        })
        return config

    @classmethod
    def from_config(cls, config):
        obj = super().from_config(config)
        obj.label_encoder = encoders.deserialize(config['encoder'])

    def fit_before_convert(self, dataset):
        """Fit the encoder."""
        # If in tf.data.Dataset, must be encoded already.
        if isinstance(dataset, tf.data.Dataset):
            return

        # Convert the data to np.ndarray.
        if isinstance(dataset, pd.DataFrame):
            dataset = dataset.values
        if isinstance(dataset, pd.Series):
            dataset = dataset.values.reshape(-1, 1)

        # If encoded.
        # TODO: support raw string labels for multi-label.
        if len(dataset.flatten()) != len(dataset):
            if self.num_classes:
                self._check_data_shape(dataset.shape[1:])
            return

        # Fit encoder.
        labels = set(dataset.flatten())
        if len(labels) < 2:
            raise ValueError('Expect the target data for {name} to have '
                             'at least 2 classes, but got {num_classes}.'
                             .format(name=self.name, num_classes=self.num_classes))
        if len(labels) == 2 and not self.multi_label:
            self.label_encoder = encoders.LabelEncoder()
        else:
            self.label_encoder = encoders.OneHotEncoder()
        self.label_encoder.fit(dataset)

    def convert_to_dataset(self, dataset):
        if self.label_encoder:
            dataset = self.label_encoder.encode(dataset)
        return super().convert_to_dataset(dataset)

    def fit(self, dataset):
        super().fit(dataset)
        shape = tuple(data_utils.dataset_shape(dataset).as_list()[1:])
        # Infer the num_classes.
        if not self.num_classes:
            # Single column with 0s and 1s.
            if shape == (1,):
                self.num_classes = 2
            else:
                self.num_classes = shape[0]
            return

        # Compute expected shape from num_classes.
        if self.num_classes == 2 and not self.multi_label:
            expected = (1,)
        else:
            expected = (self.num_classes,)

        # Check shape equals expected shape.
        if shape != expected:
            raise ValueError('Expect the target data for {name} to have '
                             'shape {expected}, but got {actual}.'
                             .format(name=self.name, expected=expected,
                                     actual=shape))

    def postprocess(self, y):
        if self.multi_label:
            y[y < 0.5] = 0
            y[y > 0.5] = 1
        if self.label_encoder:
            y = self.label_encoder.decode(y)
        return y


class RegressionHeadAdapter(HeadAdapter):
    pass


class SegmentationHeadAdapter(ClassificationHeadAdapter):
    pass



from autokeras.tuners.bayesian_optimization import BayesianOptimization
from autokeras.tuners.greedy import Greedy
from autokeras.tuners.hyperband import Hyperband
from autokeras.tuners.random_search import RandomSearch
from autokeras.tuners.task_specific import ImageClassifierTuner



import kerastuner
import tensorflow as tf
from tensorflow.python.util import nest

from autokeras import blocks as blocks_module
from autokeras import nodes as nodes_module
from autokeras.engine import head as head_module
from autokeras.engine import serializable
from autokeras.utils import utils


def feature_encoding_input(block):
    """Fetch the column_types and column_names.

    The values are fetched for FeatureEncoding from StructuredDataInput.
    """
    if not isinstance(block.inputs[0], nodes_module.StructuredDataInput):
        raise TypeError('FeatureEncoding block can only be used '
                        'with StructuredDataInput.')
    block.column_types = block.inputs[0].column_types
    block.column_names = block.inputs[0].column_names


# Compile the graph.
COMPILE_FUNCTIONS = {
    blocks_module.StructuredDataBlock: [feature_encoding_input],
    blocks_module.CategoricalToNumerical: [feature_encoding_input],
}


def load_graph(filepath, custom_objects=None):
    if custom_objects is None:
        custom_objects = {}
    with tf.keras.utils.custom_object_scope(custom_objects):
        return Graph.from_config(utils.load_json(filepath))


class Graph(kerastuner.HyperModel, serializable.Serializable):
    """A graph consists of connected Blocks, or Heads.

    # Arguments
        inputs: A list of input node(s) for the Graph.
        outputs: A list of output node(s) for the Graph.
        override_hps: A list of HyperParameters. The predefined HyperParameters that
            will override the space of the Hyperparameters defined in the Hypermodels
            with the same names.
    """

    def __init__(self, inputs=None, outputs=None, override_hps=None):
        super().__init__()
        self.inputs = nest.flatten(inputs)
        self.outputs = nest.flatten(outputs)
        self._node_to_id = {}
        self._nodes = []
        self.blocks = []
        self._block_to_id = {}
        if inputs and outputs:
            self._build_network()
        self.override_hps = override_hps or []

    def compile(self):
        """Share the information between blocks."""
        for block in self.blocks:
            for func in COMPILE_FUNCTIONS.get(block.__class__, []):
                func(block)

    def _register_hps(self, hp):
        """Register the override HyperParameters for current HyperParameters."""
        for single_hp in self.override_hps:
            name = single_hp.name
            if name not in hp.values:
                hp._register(single_hp)
                hp.values[name] = single_hp.default

    def _build_network(self):
        self._node_to_id = {}

        # Recursively find all the interested nodes.
        for input_node in self.inputs:
            self._search_network(input_node, self.outputs, set(), set())
        self._nodes = sorted(list(self._node_to_id.keys()),
                             key=lambda x: self._node_to_id[x])

        for node in (self.inputs + self.outputs):
            if node not in self._node_to_id:
                raise ValueError('Inputs and outputs not connected.')

        # Find the blocks.
        blocks = []
        for input_node in self._nodes:
            for block in input_node.out_blocks:
                if any([output_node in self._node_to_id
                        for output_node in block.outputs]) and block not in blocks:
                    blocks.append(block)

        # Check if all the inputs of the blocks are set as inputs.
        for block in blocks:
            for input_node in block.inputs:
                if input_node not in self._node_to_id:
                    raise ValueError('A required input is missing for HyperModel '
                                     '{name}.'.format(name=block.name))

        # Calculate the in degree of all the nodes
        in_degree = [0] * len(self._nodes)
        for node_id, node in enumerate(self._nodes):
            in_degree[node_id] = len([
                block for block in node.in_blocks if block in blocks])

        # Add the blocks in topological order.
        self.blocks = []
        self._block_to_id = {}
        while len(blocks) != 0:
            new_added = []

            # Collect blocks with in degree 0.
            for block in blocks:
                if any([in_degree[self._node_to_id[node]]
                        for node in block.inputs]):
                    continue
                new_added.append(block)

            # Remove the collected blocks from blocks.
            for block in new_added:
                blocks.remove(block)

            for block in new_added:
                # Add the collected blocks to the Graph.
                self._add_block(block)

                # Decrease the in degree of the output nodes.
                for output_node in block.outputs:
                    if output_node not in self._node_to_id:
                        continue
                    output_node_id = self._node_to_id[output_node]
                    in_degree[output_node_id] -= 1

    def _search_network(self, input_node, outputs, in_stack_nodes,
                        visited_nodes):
        visited_nodes.add(input_node)
        in_stack_nodes.add(input_node)

        outputs_reached = False
        if input_node in outputs:
            outputs_reached = True

        for block in input_node.out_blocks:
            for output_node in block.outputs:
                if output_node in in_stack_nodes:
                    raise ValueError('The network has a cycle.')
                if output_node not in visited_nodes:
                    self._search_network(output_node, outputs, in_stack_nodes,
                                         visited_nodes)
                if output_node in self._node_to_id.keys():
                    outputs_reached = True

        if outputs_reached:
            self._add_node(input_node)

        in_stack_nodes.remove(input_node)

    def _add_block(self, block):
        if block not in self.blocks:
            block_id = len(self.blocks)
            self._block_to_id[block] = block_id
            self.blocks.append(block)

    def _add_node(self, input_node):
        if input_node not in self._node_to_id:
            self._node_to_id[input_node] = len(self._node_to_id)

    def _get_block(self, name):
        for block in self.blocks:
            if block.name == name:
                return block
        raise ValueError('Cannot find block named {name}.'.format(name=name))

    def get_config(self):
        blocks = [blocks_module.serialize(block) for block in self.blocks]
        nodes = {str(self._node_to_id[node]): nodes_module.serialize(node)
                 for node in self.inputs}
        override_hps = [kerastuner.engine.hyperparameters.serialize(hp)
                        for hp in self.override_hps]
        block_inputs = {
            str(block_id): [self._node_to_id[node]
                            for node in block.inputs]
            for block_id, block in enumerate(self.blocks)}
        block_outputs = {
            str(block_id): [self._node_to_id[node]
                            for node in block.outputs]
            for block_id, block in enumerate(self.blocks)}

        outputs = [self._node_to_id[node] for node in self.outputs]

        return {
            'override_hps': override_hps,  # List [serialized].
            'blocks': blocks,  # Dict {id: serialized}.
            'nodes': nodes,  # Dict {id: serialized}.
            'outputs': outputs,  # List of node_ids.
            'block_inputs': block_inputs,  # Dict {id: List of node_ids}.
            'block_outputs': block_outputs,  # Dict {id: List of node_ids}.
        }

    @classmethod
    def from_config(cls, config):
        blocks = [blocks_module.deserialize(block) for block in config['blocks']]
        nodes = {int(node_id): nodes_module.deserialize(node)
                 for node_id, node in config['nodes'].items()}
        override_hps = [kerastuner.engine.hyperparameters.deserialize(config)
                        for config in config['override_hps']]

        inputs = [nodes[node_id] for node_id in nodes]
        for block_id, block in enumerate(blocks):
            input_nodes = [nodes[node_id]
                           for node_id in config['block_inputs'][str(block_id)]]
            output_nodes = nest.flatten(block(input_nodes))
            for output_node, node_id in zip(
                    output_nodes, config['block_outputs'][str(block_id)]):
                nodes[node_id] = output_node

        outputs = [nodes[node_id] for node_id in config['outputs']]
        return cls(inputs=inputs, outputs=outputs, override_hps=override_hps)

    def build(self, hp):
        """Build the HyperModel into a Keras Model."""
        tf.keras.backend.clear_session()
        self._register_hps(hp)
        self.compile()
        real_nodes = {}
        for input_node in self.inputs:
            node_id = self._node_to_id[input_node]
            real_nodes[node_id] = input_node.build()
        for block in self.blocks:
            temp_inputs = [real_nodes[self._node_to_id[input_node]]
                           for input_node in block.inputs]
            outputs = block.build(hp, inputs=temp_inputs)
            outputs = nest.flatten(outputs)
            for output_node, real_output_node in zip(block.outputs, outputs):
                real_nodes[self._node_to_id[output_node]] = real_output_node
        model = tf.keras.Model(
            [real_nodes[self._node_to_id[input_node]] for input_node in
             self.inputs],
            [real_nodes[self._node_to_id[output_node]] for output_node in
             self.outputs])

        return self._compile_keras_model(hp, model)

    def _get_metrics(self):
        metrics = {}
        for output_node in self.outputs:
            block = output_node.in_blocks[0]
            if isinstance(block, head_module.Head):
                metrics[block.name] = block.metrics
        return metrics

    def _get_loss(self):
        loss = {}
        for output_node in self.outputs:
            block = output_node.in_blocks[0]
            if isinstance(block, head_module.Head):
                loss[block.name] = block.loss
        return loss

    def _compile_keras_model(self, hp, model):
        # Specify hyperparameters from compile(...)
        optimizer_name = hp.Choice('optimizer',
                                   ['adam', 'adadelta', 'sgd'],
                                   default='adam')
        learning_rate = hp.Choice('learning_rate',
                                  [1e-1, 1e-2, 1e-3, 1e-4, 1e-5],
                                  default=1e-3)

        if optimizer_name == 'adam':
            optimizer = tf.keras.optimizers.Adam(
                learning_rate=learning_rate)
        elif optimizer_name == 'adadelta':
            optimizer = tf.keras.optimizers.Adadelta(
                learning_rate=learning_rate)
        elif optimizer_name == 'sgd':
            optimizer = tf.keras.optimizers.SGD(
                learning_rate=learning_rate)

        model.compile(optimizer=optimizer,
                      metrics=self._get_metrics(),
                      loss=self._get_loss())

        return model

    def save(self, filepath):
        utils.save_json(filepath, self.get_config())



import numpy as np
import pytest
import tensorflow as tf

from autokeras.adapters import output_adapter
from tests import utils


def test_y_is_pd_series():
    (x, y), (val_x, val_y) = utils.dataframe_series()
    adapter = output_adapter.ClassificationHeadAdapter(name='a')
    adapter.fit_transform(y)
    assert isinstance(adapter.transform(y), tf.data.Dataset)


def test_unsupported_types():
    y = 1
    adapter = output_adapter.ClassificationHeadAdapter(name='a')
    with pytest.raises(TypeError) as info:
        adapter.check(y)
    assert 'Expect the target data' in str(info.value)


def test_one_class():
    y = np.array(['a', 'a', 'a'])
    adapter = output_adapter.ClassificationHeadAdapter(name='a')
    with pytest.raises(ValueError) as info:
        adapter.fit_before_convert(y)
    assert 'Expect the target data' in str(info.value)


def test_infer_num_classes():
    y = utils.generate_one_hot_labels(dtype='dataset')
    adapter = output_adapter.ClassificationHeadAdapter(name='a')
    y = adapter.fit(y)
    assert adapter.num_classes == 10


def test_infer_two_classes():
    y = tf.data.Dataset.from_tensor_slices(np.random.rand(10, 1)).batch(32)
    adapter = output_adapter.ClassificationHeadAdapter(name='a')
    y = adapter.fit(y)
    assert adapter.num_classes == 2


def test_check_data_shape():
    y = tf.data.Dataset.from_tensor_slices(np.random.rand(10, 5)).batch(32)
    adapter = output_adapter.ClassificationHeadAdapter(name='a', num_classes=5)
    adapter.fit(y)


def test_check_data_shape_two_classes():
    y = tf.data.Dataset.from_tensor_slices(np.random.rand(10, 1)).batch(32)
    adapter = output_adapter.ClassificationHeadAdapter(name='a', num_classes=2)
    adapter.fit(y)


def test_check_data_shape_error():
    y = tf.data.Dataset.from_tensor_slices(np.random.rand(10, 3)).batch(32)
    adapter = output_adapter.ClassificationHeadAdapter(name='a', num_classes=5)
    with pytest.raises(ValueError) as info:
        adapter.fit(y)
    assert 'Expect the target data for a to have shape' in str(info.value)


def test_multi_label_two_classes():
    y = np.random.rand(10, 2)
    adapter = output_adapter.ClassificationHeadAdapter(name='a', multi_label=True)
    adapter.fit_transform(y)
    assert adapter.label_encoder is None


def test_multi_label_postprocessing():
    y = np.random.rand(10, 3)
    adapter = output_adapter.ClassificationHeadAdapter(name='a', multi_label=True)
    adapter.fit_transform(y)
    y = adapter.postprocess(y)
    assert set(y.flatten().tolist()) == set([1, 0])



import tensorflow as tf

from autokeras.engine import block as block_module
from autokeras.engine import io_hypermodel


def serialize_metrics(metrics):
    serialized = []
    for metric in metrics:
        if isinstance(metric, str):
            serialized.append([metric])
        else:
            serialized.append(tf.keras.metrics.serialize(metric))
    return serialized


def deserialize_metrics(metrics):
    deserialized = []
    for metric in metrics:
        if isinstance(metric, list):
            deserialized.append(metric[0])
        else:
            deserialized.append(tf.keras.metrics.deserialize(metric))
    return deserialized


def serialize_loss(loss):
    if isinstance(loss, str):
        return [loss]
    return tf.keras.losses.serialize(loss)


def deserialize_loss(loss):
    if isinstance(loss, list):
        return loss[0]
    return tf.keras.losses.deserialize(loss)


class Head(block_module.Block, io_hypermodel.IOHyperModel):
    """Base class for the heads, e.g. classification, regression.

    # Arguments
        loss: A Keras loss function. Defaults to None. If None, the loss will be
            inferred from the AutoModel.
        metrics: A list of Keras metrics. Defaults to None. If None, the metrics will
            be inferred from the AutoModel.
        output_shape: Tuple of int(s). Defaults to None. If None, the output shape
            will be inferred from the AutoModel.
    """

    def __init__(self, loss=None, metrics=None, output_shape=None, **kwargs):
        super().__init__(**kwargs)
        self.output_shape = output_shape
        self.loss = loss
        if metrics is None:
            metrics = []
        self.metrics = metrics
        # Mark if the head should directly output the input tensor.

    def get_config(self):
        config = super().get_config()
        config.update({
            'loss': serialize_loss(self.loss),
            'metrics': serialize_metrics(self.metrics),
            'output_shape': self.output_shape
        })
        return config

    @classmethod
    def from_config(cls, config):
        config['loss'] = deserialize_loss(config['loss'])
        config['metrics'] = deserialize_metrics(config['metrics'])
        return super().from_config(config)

    def build(self, hp, inputs=None):
        raise NotImplementedError

    def config_from_adapter(self, adapter):
        self.output_shape = adapter.shape



from typedapi import ensure_api_is_typed

import autokeras

HELP_MESSAGE = (
    "You can also take a look at this issue:\n"
    "https://github.com/keras-team/autokeras/issues/918"
)


# TODO: add types and remove all elements from
# the exception list.
EXCEPTION_LIST = [
    autokeras.Head,
    autokeras.Node,
    autokeras.CategoricalToNumerical,
    autokeras.Flatten,
    autokeras.ImageAugmentation,
    autokeras.ImageBlock,
    autokeras.StructuredDataBlock,
    autokeras.TextBlock,
    autokeras.TextToNgramVector,
    autokeras.ImageInput,
    autokeras.Input,
    autokeras.StructuredDataInput,
    autokeras.TimeseriesInput,
    autokeras.TextInput,
    autokeras.ImageClassifier,
    autokeras.ImageRegressor,
    autokeras.StructuredDataClassifier,
    autokeras.TextRegressor,
    autokeras.TimeseriesForecaster,
    autokeras.BayesianOptimization,
    autokeras.Greedy,
    autokeras.Hyperband,
    autokeras.RandomSearch,
]


def test_api_surface_is_typed():
    ensure_api_is_typed(
        [autokeras], EXCEPTION_LIST, init_only=True, additional_message=HELP_MESSAGE,
    )



import kerastuner

from autokeras.engine import tuner as tuner_module


class Hyperband(kerastuner.Hyperband, tuner_module.AutoTuner):
    """KerasTuner Hyperband with preprocessing layer tuning."""

    def __init__(self, max_epochs=1000, max_trials=100, *args, **kwargs):
        super().__init__(max_epochs=max_epochs, *args, **kwargs)
        self.oracle.max_trials = max_trials



import tensorflow as tf

import autokeras as ak
from autokeras import nodes


def test_time_series_input_node():
    # TODO. Change test once TimeSeriesBlock is added.
    node = ak.TimeseriesInput(shape=(32,), lookback=2)
    output = node.build()
    assert isinstance(output, tf.Tensor)

    node = nodes.deserialize(nodes.serialize(node))
    output = node.build()
    assert isinstance(output, tf.Tensor)



import json
import re

import kerastuner
import tensorflow as tf
from packaging.version import parse
from tensorflow.python.util import nest


def validate_num_inputs(inputs, num):
    inputs = nest.flatten(inputs)
    if not len(inputs) == num:
        raise ValueError('Expected {num} elements in the inputs list '
                         'but received {len} inputs.'.format(num=num,
                                                             len=len(inputs)))


def get_name_scope():
    with tf.name_scope('a') as scope:
        name_scope = scope[:-2]
    return name_scope


def to_snake_case(name):
    intermediate = re.sub('(.)([A-Z][a-z0-9]+)', r'\1_\2', name)
    insecure = re.sub('([a-z])([A-Z])', r'\1_\2', intermediate).lower()
    # If the class is private the name starts with "_" which is not secure
    # for creating scopes. We prefix the name with "private" in this case.
    if insecure[0] != '_':
        return insecure
    return 'private' + insecure


def to_type_key(dictionary, convert_func):
    """Convert the keys of a dictionary to a different type.

    # Arguments
        dictionary: Dictionary. The dictionary to be converted.
        convert_func: Function. The function to convert a key.
    """
    return {convert_func(key): value
            for key, value in dictionary.items()}


def check_tf_version() -> None:
    if parse(tf.__version__) < parse('2.2.0'):
        raise ImportError(
            'The Tensorflow package version needs to be at least 2.2.0 \n'
            'for AutoKeras to run. Currently, your TensorFlow version is \n'
            '{version}. Please upgrade with \n'
            '`$ pip install --upgrade tensorflow`. \n'
            'You can use `pip freeze` to check afterwards that everything is '
            'ok.'.format(version=tf.__version__)
        )


def check_kt_version() -> None:
    if parse(kerastuner.__version__) < parse('1.0.2rc0'):
        raise ImportError(
            'The Keras Tuner package version needs to be at least 1.0.2rc0 \n'
            'for AutoKeras to run. Currently, your Keras Tuner version is \n'
            '{version}. Please upgrade with \n'
            '`$ pip install '
            'git+https://github.com/keras-team/keras-tuner.git@1.0.2rc0'
            '#egg=keras-tuner-1.0.2rc0`. \n'
            'You can use `pip freeze` to check afterwards that everything is '
            'ok.'.format(version=kerastuner.__version__)
        )


def save_json(path, obj):
    obj = json.dumps(obj)
    with tf.io.gfile.GFile(path, 'w') as f:
        f.write(obj)


def load_json(path):
    with tf.io.gfile.GFile(path, 'r') as f:
        obj = f.read()
    return json.loads(obj)


def contain_instance(instance_list, instance_type):
    return any([isinstance(instance, instance_type)
                for instance in instance_list])



import inspect

import tensorflow as tf
from tensorflow.keras.layers.experimental import preprocessing
from tensorflow.python.keras.layers.preprocessing import index_lookup
from tensorflow.python.util import nest

CombinerPreprocessingLayer = inspect.getmro(preprocessing.Normalization)[1]
Combiner = inspect.getmro(preprocessing.Normalization()._combiner.__class__)[1]

INT = 'int'
NONE = 'none'
ONE_HOT = 'one-hot'


class MultiColumnCategoricalEncoding(preprocessing.PreprocessingLayer):
    """Encode the categorical features to numerical features.

    # Arguments
        encoding: A list of strings, which has the same number of elements as the
            columns in the structured data. Each of the strings specifies the
            encoding method used for the corresponding column. Use 'int' for
            categorical columns and 'none' for numerical columns.
    """

    # TODO: Support one-hot encoding.
    # TODO: Support frequency encoding.

    def __init__(self, encoding, **kwargs):
        super().__init__(**kwargs)
        self.encoding = encoding
        self.encoding_layers = []
        for encoding in self.encoding:
            if encoding == NONE:
                self.encoding_layers.append(None)
            elif encoding == INT:
                self.encoding_layers.append(index_lookup.IndexLookup())
            elif encoding == ONE_HOT:
                self.encoding_layers.append(None)

    def build(self, input_shape):
        for encoding_layer in self.encoding_layers:
            if encoding_layer is not None:
                encoding_layer.build(tf.TensorShape([1]))

    def call(self, inputs):
        input_nodes = nest.flatten(inputs)[0]
        split_inputs = tf.split(input_nodes, [1] * len(self.encoding), axis=-1)
        output_nodes = []
        for input_node, encoding_layer in zip(split_inputs, self.encoding_layers):
            if encoding_layer is None:
                number = tf.strings.to_number(input_node, tf.float32)
                # Replace NaN with 0.
                imputed = tf.where(tf.math.is_nan(number),
                                   tf.zeros_like(number),
                                   number)
                output_nodes.append(imputed)
            else:
                output_nodes.append(tf.cast(encoding_layer(input_node), tf.float32))
        if len(output_nodes) == 1:
            return output_nodes[0]
        return tf.keras.layers.Concatenate()(output_nodes)

    def adapt(self, data):
        for index, encoding_layer in enumerate(self.encoding_layers):
            if encoding_layer is None:
                continue
            data_column = data.map(lambda x: tf.slice(x, [0, index], [-1, 1]))
            encoding_layer.adapt(data_column)

    def get_config(self):
        config = {
            'encoding': self.encoding,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


CUSTOM_OBJECTS = {
    'MultiColumnCategoricalEncoding': MultiColumnCategoricalEncoding,
    'IndexLookup': index_lookup.IndexLookup,
}



from tensorflow.python.util import nest

from autokeras.blocks import basic
from autokeras.blocks import preprocessing
from autokeras.blocks import reduction
from autokeras.engine import block as block_module


class ImageBlock(block_module.Block):
    """Block for image data.

    The image blocks is a block choosing from ResNetBlock, XceptionBlock, ConvBlock,
    which is controlled by a hyperparameter, 'block_type'.

    # Arguments
        block_type: String. 'resnet', 'xception', 'vanilla'. The type of Block
            to use. If unspecified, it will be tuned automatically.
        normalize: Boolean. Whether to channel-wise normalize the images.
            If unspecified, it will be tuned automatically.
        augment: Boolean. Whether to do image augmentation. If unspecified,
            it will be tuned automatically.
    """

    def __init__(self,
                 block_type=None,
                 normalize=None,
                 augment=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.block_type = block_type
        self.normalize = normalize
        self.augment = augment

    def get_config(self):
        config = super().get_config()
        config.update({'block_type': self.block_type,
                       'normalize': self.normalize,
                       'augment': self.augment})
        return config

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        output_node = input_node

        block_type = self.block_type or hp.Choice('block_type',
                                                  ['resnet', 'xception', 'vanilla'],
                                                  default='vanilla')

        normalize = self.normalize
        if normalize is None:
            normalize = hp.Boolean('normalize', default=False)
        augment = self.augment
        if augment is None:
            augment = hp.Boolean('augment', default=False)
        if normalize:
            output_node = preprocessing.Normalization().build(hp, output_node)
        if augment:
            output_node = preprocessing.ImageAugmentation().build(hp, output_node)
        if block_type == 'resnet':
            output_node = basic.ResNetBlock().build(hp, output_node)
        elif block_type == 'xception':
            output_node = basic.XceptionBlock().build(hp, output_node)
        elif block_type == 'vanilla':
            output_node = basic.ConvBlock().build(hp, output_node)
        return output_node


class TextBlock(block_module.Block):
    """Block for text data.

    # Arguments
        block_type: String. 'vanilla', 'transformer', and 'ngram'. The type of Block
            to use. 'vanilla' and 'transformer' use a TextToIntSequence vectorizer,
            whereas 'ngram' uses TextToNgramVector. If unspecified, it will be tuned
            automatically.
        max_tokens: Int. The maximum size of the vocabulary.
            If left unspecified, it will be tuned automatically.
        pretraining: String. 'random' (use random weights instead any pretrained
            model), 'glove', 'fasttext' or 'word2vec'. Use pretrained word embedding.
            If left unspecified, it will be tuned automatically.
    """

    def __init__(self,
                 block_type=None,
                 max_tokens=None,
                 pretraining=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.block_type = block_type
        self.max_tokens = max_tokens
        self.pretraining = pretraining

    def get_config(self):
        config = super().get_config()
        config.update({
            'block_type': self.block_type,
            'max_tokens': self.max_tokens,
            'pretraining': self.pretraining})
        return config

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        output_node = input_node
        block_type = self.block_type or hp.Choice('block_type',
                                                  ['vanilla',
                                                   'transformer',
                                                   'ngram'],
                                                  default='vanilla')
        max_tokens = self.max_tokens or hp.Choice('max_tokens',
                                                  [500, 5000, 20000],
                                                  default=5000)
        if block_type == 'ngram':
            output_node = preprocessing.TextToNgramVector(
                max_tokens=max_tokens).build(hp, output_node)
            output_node = basic.DenseBlock().build(hp, output_node)
        else:
            output_node = preprocessing.TextToIntSequence(
                max_tokens=max_tokens).build(hp, output_node)
            if block_type == 'transformer':
                output_node = basic.Transformer(max_features=max_tokens + 1,
                                                pretraining=self.pretraining
                                                ).build(hp, output_node)
            else:
                output_node = basic.Embedding(
                    max_features=max_tokens + 1,
                    pretraining=self.pretraining).build(hp, output_node)
                output_node = basic.ConvBlock().build(hp, output_node)
            output_node = reduction.SpatialReduction().build(hp, output_node)
            output_node = basic.DenseBlock().build(hp, output_node)
        return output_node


class StructuredDataBlock(block_module.Block):
    """Block for structured data.

    # Arguments
        categorical_encoding: Boolean. Whether to use the CategoricalToNumerical to
            encode the categorical features to numerical features. Defaults to True.
            If specified as None, it will be tuned automatically.
        seed: Int. Random seed.
    """

    def __init__(self,
                 categorical_encoding=True,
                 seed=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.categorical_encoding = categorical_encoding
        self.seed = seed
        self.column_types = None
        self.column_names = None

    @classmethod
    def from_config(cls, config):
        column_types = config.pop('column_types')
        column_names = config.pop('column_names')
        instance = cls(**config)
        instance.column_types = column_types
        instance.column_names = column_names
        return instance

    def get_config(self):
        config = super().get_config()
        config.update({'categorical_encoding': self.categorical_encoding,
                       'seed': self.seed,
                       'column_types': self.column_types,
                       'column_names': self.column_names})
        return config

    def build_categorical_encoding(self, hp, input_node):
        output_node = input_node
        categorical_encoding = self.categorical_encoding
        if categorical_encoding is None:
            categorical_encoding = hp.Choice('categorical_encoding',
                                             [True, False],
                                             default=True)
        if categorical_encoding:
            block = preprocessing.CategoricalToNumerical()
            block.column_types = self.column_types
            block.column_names = self.column_names
            output_node = block.build(hp, output_node)
        return output_node

    def build_body(self, hp, input_node):
        output_node = basic.DenseBlock().build(hp, input_node)
        return output_node

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        output_node = self.build_categorical_encoding(hp, input_node)
        output_node = self.build_body(hp, output_node)
        return output_node


class TimeseriesBlock(block_module.Block):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_config(self):
        config = super().get_config()
        return config

    def build(self, hp, inputs=None):
        input_node = nest.flatten(inputs)[0]
        output_node = input_node
        output_node = basic.RNNBlock().build(hp, output_node)
        return output_node


class GeneralBlock(block_module.Block):
    """A general neural network block when the input type is unknown.

    When the input type is unknown. The GeneralBlock would search in a large space
    for a good model.

    # Arguments
        name: String.
    """

    def build(self, hp, inputs=None):
        raise NotImplementedError
